#!/usr/bin/env python
#
# main.py - funpack entry point
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the ``funpack`` entry point. """


import multiprocessing     as mp
import                        sys
import                        logging
import                        fnmatch
import                        warnings
import                        datetime
import                        calendar

import pandas              as pd

import funpack
import funpack.util        as util
import funpack.icd10       as icd10
import funpack.config      as config
import funpack.custom      as custom
import funpack.dryrun      as dryrun
import funpack.cleaning    as cleaning
import funpack.importing   as importing
import funpack.exporting   as exporting
import funpack.hierarchy   as hierarchy
import funpack.processing  as processing
import funpack.loadtables  as loadtables


log = logging.getLogger(__name__)


def main(argv=None):
    """``funpack`` entry point. """

    # Make sure built in plugins are
    # registered, as they are queried
    # in the command-line help. Set
    # logging to critical until we've
    # parsed command-line args.
    logging.getLogger().setLevel(logging.CRITICAL)
    custom.registerBuiltIns()

    args, argv = config.parseArgsWithConfigFile(argv)
    date = datetime.date.today()

    # Now that args are passed,
    # we can set up logging properly.
    configLogging(args)

    if bool(getattr(args, 'low_memory', None)):
        warnings.warn('The --low_memory option is deprecated '
                      'and has no effect.', DeprecationWarning)
    if bool(getattr(args, 'work_dir', None)):
        warnings.warn('The --work_dir option is deprecated '
                      'and has no effect.', DeprecationWarning)

    log.info('funpack %s', funpack.__version__)
    log.info('Date: %s (%s)', date.today(), calendar.day_name[date.weekday()])
    log.info('Command-line arguments %s', ' '.join(argv))
    log.debug('Running with the following options')
    for name, val in args.__dict__.items():
        if val is not None:
            val = str(val)
            if len(val) <= 30: log.debug('  %s: %s',    name, val)
            else:              log.debug('  %s: %s...', name, val[:30])

    # Re-load any custom plugins
    # that have been specified.
    custom.registerBuiltIns()

    if args.plugin_file is not None:
        for p in args.plugin_file:
            custom.loadPluginFile(p)

    # error if any loaders/formats are
    # invalid (we can only perform this
    # check after plugins have been
    # loaded)
    if args.loader is not None:
        for f, l in args.loader.items():
            if not custom.exists('loader', l):
                raise ValueError('Unknown loader {} [{}]'.format(l, f))
    if not custom.exists('exporter', args.format):
        raise ValueError('Unknown output format {}'.format(args.format))
    if args.date_format is not None and \
       not custom.exists('formatter', args.date_format):
        raise ValueError('Unknown date format {}'.format(args.date_format))
    if args.time_format is not None and \
       not custom.exists('formatter', args.time_format):
        raise ValueError('Unknown time format {}'.format(args.time_format))
    if args.tsv_var_format is not None:
        for v, f in args.tsv_var_format.items():
            if not custom.exists('formatter', f):
                raise ValueError('Unknown formatter {} [{}]'.format(f, v))

    if args.num_jobs > 1:
        log.debug('Running up to %i jobs in parallel', args.num_jobs)
        mgr = mp.Manager()

        # We need to initialise icd10
        # before the worker processes
        # are created, so its state is
        # shared by all processes.
        icd10.initialise(mgr)

        pool = mp.Pool(args.num_jobs)

    else:
        pool = None
        mgr  = None

    try:
        with util.timed(None, log, fmt='Total time: %s (%+iMB)'):

            dtable, unknowns, unprocessed, drop = doImport(args, pool, mgr)

            if args.dry_run:
                dryrun.doDryRun(dtable, unknowns, unprocessed, drop, args)
            else:
                doCleanAndProcess(  dtable, args)
                allcols = list(dtable.dataColumns)
                finaliseColumns(    dtable, args, unknowns)
                doUnknownsExport(   dtable, args, unknowns, unprocessed,
                                    allcols)
                doExport(           dtable, args)
                doICD10Export(              args)
                doDescriptionExport(dtable, args)
                doSummaryExport(    dtable, args)

    finally:
        # shutdown the pool gracefully
        if pool is not None:
            pool.close()
            pool.join()
            pool = None

    return 0


def doImport(args, pool, mgr):
    """Data import stage.

    :arg args: :class:`argparse.Namespace` object containing command line
               arguments
    :arg pool: :class:`multiprocessing.Pool` object for parallelisation (may
               be ``None``)
    :arg mgr:  :class:`multiprocessing.Manager` object for parallelisation (may
               be ``None``)

    :returns:  A tuple containing:

                - A :class:`.DataTable` containing the data
                - A sequence of :class:`.Column` objects representing the
                  unknown columns.
                - A sequence of :class:`.Column` objects representing columns
                  which are uncategorised, and have no processing or cleaning
                  rules specified on them.
    """

    with util.timed('Table import', log):
        vartable, proctable, cattable, unknowns, unprocessed = \
            loadtables.loadTables(
                args.infile,
                args.variable_file,
                args.datacoding_file,
                args.type_file,
                args.processing_file,
                args.category_file,
                noBuiltins=args.no_builtins,
                naValues=args.na_values,
                childValues=args.child_values,
                recoding=args.recoding,
                clean=args.clean,
                typeClean=args.type_clean,
                globalClean=args.global_clean,
                skipProcessing=args.skip_processing,
                prependProcess=args.prepend_process,
                appendProcess=args.append_process,
                sniffers=args.loader,
                indexes=args.index)

        # if it appears that we're doing a
        # full run on a large data set, emit
        # warnings about unknown/unprocessed
        # variables.
        bigrun = any((args.variable_file   is not None,
                      args.datacoding_file is not None,
                      args.processing_file is not None,
                      args.category_file   is not None))

        if bigrun:
            for u in unknowns:
                log.warning('Variable %s [file %s, column %s, assigned '
                            'variable ID %s] is unknown.',
                            u.name, u.datafile, u.index, u.vid)
            for u in unprocessed:
                log.warning('Variable %s [file %s, column %s, assigned '
                            'variable ID %s] is uncategorised and does not '
                            'have any cleaning or processing rules set.',
                            u.name, u.datafile, u.index, u.vid)

    subjects, exprs = args.subject

    if not args.dry_run and args.import_all:
        variables     = None
        categories    = None
        columns       = None
        removeUnknown = None
    else:
        variables     = args.variable
        categories    = args.category
        columns       = args.column
        removeUnknown = args.remove_unknown

    # Import data
    with util.timed('Data import', log):
        dtable, drop = importing.importData(
            datafiles=args.infile,
            vartable=vartable,
            proctable=proctable,
            cattable=cattable,
            variables=variables,
            colnames=columns,
            categories=categories,
            subjects=subjects,
            subjectExprs=exprs,
            exclude=args.exclude,
            encoding=args.encoding,
            indexes=args.index,
            unknownVars=unknowns,
            trustTypes=args.trust_types,
            removeUnknown=removeUnknown,
            mergeAxis=args.merge_axis,
            mergeStrategy=args.merge_strategy,
            indexVisits=args.index_visits,
            loaders=args.loader,
            pool=pool,
            mgr=mgr,
            dryrun=args.dry_run)

    return dtable, unknowns, unprocessed, drop


def doCleanAndProcess(dtable, args):
    """Data cleaning and processing stage.

    :arg dtable: :class:`.DataTable` containing the data
    :arg args:   :class:`argparse.Namespace` object containing command line
                 arguments
    :arg pool:   :class:`multiprocessing.Pool` object for parallelisation (may
                 be ``None``)
    """

    # Clean data (it times each step individually)
    cleaning.cleanData(
        dtable,
        skipNAInsertion=args.skip_insertna,
        skipCleanFuncs=args.skip_clean_funcs,
        skipChildValues=args.skip_childvalues,
        skipRecoding=args.skip_recoding)

    # Process data
    with util.timed('Data processing', log):
        processing.processData(dtable)


def finaliseColumns(dtable, args, unknowns):
    """Called after processing and before export.

    If the ``--import_all`` argument was used (which forces all columns to be
    loaded and processed), this function applies the ``--variable``,
    ``--category``, ``--column``, and ``--remove_unknown`` arguments. to the
    processed data.

    :arg dtable:      :class:`.DataTable` containing the data
    :arg args:        :class:`argparse.Namespace` object containing command
                      line arguments
    :arg unknowns:    List of :class:`.Column` objects representing the
                      unknown columns.
    :arg unprocessed: A sequence of :class:`.Column` objects representing
                      columns which are uncategorised, and have no processing
                      or cleaning rules specified on them.
    """

    if not args.import_all:
        return

    # get a list of variables requested
    # via --variable or --category (will
    # be None if no requests)
    vids = importing.restrictVariables(
        dtable.cattable, args.variable, args.category)

    # args.remove_unknown is only applied
    # if variables/columns were not already
    # restricted by args.variable,
    # args.category, and/or args.column
    removeUnknown = all((vids is None,
                         args.remove_unknown,
                         args.column is None,
                         len(unknowns) > 0))

    # Build a final list of variable
    # IDs to be exported
    finalvids  = None
    removecols = []

    # apply removeUnknown - again, this
    # is only relevant if no variables/
    # categories/columns were specified
    if removeUnknown:
        finalvids  = dtable.variables
        unkvids    = {c.vid for c in unknowns}
        for vid in list(finalvids):
            if vid in unkvids:
                finalvids.remove(vid)

    # filter based on requested
    # variables and column names
    elif (vids is not None) or (args.column is not None):

        if vids is not None: finalvids = vids
        else:                finalvids = []

        # figure out vids of requested
        # column names/patterns
        colvids = []
        if args.column is not None:
            for col in dtable.dataColumns:

                # this VID already requested -
                # column pattern irrelevant
                if col.vid in finalvids:
                    continue
                hits = [fnmatch.fnmatch(col.name, pat) for pat in args.column]

                # include/exclude this column
                # based on whether it matches
                # any column names/patterns
                if any(hits):
                    if col.vid not in colvids:
                        colvids.append(col.vid)
                elif col.vid in colvids:
                    removecols.append(col)

            # columns requested by column pattern
            # are inserted at the beginning
            finalvids = colvids + finalvids

    # reorder variables
    if finalvids is not None:
        log.debug('Re-ordering columns for export')
        finalvids = [vid for vid in finalvids if dtable.present(vid)]
        dtable.order(finalvids)

    # remove columns
    if len(removecols) > 0:
        log.debug('Removing %u columns according to name patterns',
                  len(removecols))
        dtable.removeColumns(removecols)


def doExport(dtable, args):
    """Data export stage.

    :arg dtable: :class:`.DataTable` containing the data
    :arg args:   :class:`argparse.Namespace` object containing command line
                 arguments
    """

    # If exporting to TSV, and not parallelising,
    # we export the entire file in one go. Because
    # what's the point in chunked export if we're
    # not parallelising across chunks?
    if args.num_jobs <= 1:
        args.num_rows = len(dtable)

    with util.timed('Data export', log):
        exporting.exportData(
            dtable,
            args.outfile,

            # General export options
            colpat=args.column_pattern,
            colmap=args.rename_column,
            idcol=args.output_id_column,
            fileFormat=args.format,
            dateFormat=args.date_format,
            timeFormat=args.time_format,
            numRows=args.num_rows,

            # TSV options
            sep=args.tsv_sep,
            missingValues=args.tsv_missing_values,
            formatters=args.tsv_var_format,
            nonNumericFile=args.non_numeric_file,

            # HDF5 options
            key=args.hdf5_key,
            style=args.hdf5_style)


def doICD10Export(args):
    """If a ``--icd10_map_file`` has been specified, the ICD10 codes present
    in the data (and their converted values) are saved out to the file.
    """
    if args.icd10_map_file is None:
        return

    with util.timed('ICD10 mapping export', log):
        try:
            ihier = hierarchy.getHierarchyFilePath(name='icd10')
            ihier = hierarchy.loadHierarchyFile(ihier)
            icd10.saveCodes(args.icd10_map_file, ihier)

        except Exception as e:
            log.warning('Failed to export ICD10 mappings: {}'.format(e),
                        exc_info=True)


def doDescriptionExport(dtable, args):
    """If a ``--description_file`` has been specified, a description for every
    column is saved out to the file.
    """
    if args.description_file is None:
        return

    with util.timed('Description export', log):
        cols = dtable.dataColumns

        try:
            with open(args.description_file, 'wt') as f:
                for col in cols:
                    desc = generateDescription(dtable, col)
                    f.write('{}\t{}\n'.format(col.name, desc))

        except Exception as e:
            log.warning('Failed to export descriptions: {}'.format(e),
                        exc_info=True)


def generateDescription(dtable, col):
    """Called by :func:`doDescriptionExport`. Generates and returns a
    suitable description for the given column.

    :arg dtable: :class:`.Datatable` instance
    :arg col:    :class:`.Column` instance
    """
    vartable = dtable.vartable
    desc     = vartable.loc[col.vid, 'Description']

    if pd.isna(desc) or (desc == col.name):
        desc = 'n/a'

    # If metadata has been added to the column,
    # we add it to the description. See the
    # binariseCategorical processing function
    # for an example of this.
    if col.metadata is not None:
        suffix = ' ({})'.format(col.metadata)
    else:
        suffix = ' ({}.{})'.format(col.visit, col.instance)

    return '{}{}'.format(desc, suffix)


def doUnknownsExport(dtable, args, unknowns, unprocessed, allcols):
    """If the ``--unknown_vars_file`` argument was used, the unknown/
    unprocessed columns are saved out to a file.

    :arg dtable:      :class:`.DataTable` containing the data
    :arg args:        :class:`argparse.Namespace` object containing command
                      line arguments
    :arg unknowns:    List of :class:`.Column` objects representing the
                      unknown columns.
    :arg unprocessed: A sequence of :class:`.Column` objects representing
                      columns which are uncategorised, and have no processing
                      or cleaning rules specified on them.
    :arg allcols:     List of :class:`.Column` objects describing all
                      columns that were present in the data before being
                      filtered by the :func:`finaliseColumns` function.
    """

    if (args.unknown_vars_file is None) or (not args.import_all):
        return

    # Save unknown/unprocessed vars list to file
    # columns:
    #  - name      - column name
    #  - file      - originating input file
    #  - class     - unknown or uncategorised/unprocessed
    #  - processed - whether column passed processing
    #  - exported  - whether column was exported
    finalcols   = list(dtable.dataColumns)
    allunknowns = list(unknowns + unprocessed)

    names       = [    u.name          for u in allunknowns]
    files       = [    u.datafile      for u in allunknowns]
    classes     = ['unknown'           for u in unknowns] + \
                  ['unprocessed'       for u in unprocessed]
    processed   = [int(u in allcols)   for u in allunknowns]
    exported    = [int(u in finalcols) for u in allunknowns]
    rows        = ['{}\t{}\t{}\t{}\t{}'.format(n, f, c, p, e)
                   for n, f, c, p, e
                   in zip(names, files, classes, processed, exported)]

    log.debug('Saving unknown/unprocessed variables to %s',
              args.unknown_vars_file)

    try:
        with open(args.unknown_vars_file, 'wt') as f:
            f.write('name\tfile\tclass\tprocessed\texported\n')
            f.write('\n'.join(rows))

    except Exception as e:
        log.warning('Error saving unknown variables to {}: '
                    '{}'.format(args.unknown_vars_file, e),
                    exc_info=True)



def doSummaryExport(dtable, args):
    """If a ``--summary_file`` has been specified, a summary of the cleaning
    steps that have been applied to each variable are saved out to the file.
    """
    if args.summary_file is None:
        return

    vartable = dtable.vartable
    vids     = sorted(dtable.variables)[1:]
    sum      = pd.DataFrame(columns=['ID', 'NAValues',
                                     'RawLevels', 'NewLevels',
                                     'ParentValues', 'ChildValues',
                                     'Clean', 'Flags']).set_index('ID')

    with util.timed('Summary export', log):
        for vid in vids:
            sum.at[vid, 'NAValues']     = vartable.at[vid, 'NAValues']
            sum.at[vid, 'RawLevels']    = vartable.at[vid, 'RawLevels']
            sum.at[vid, 'NewLevels']    = vartable.at[vid, 'NewLevels']
            sum.at[vid, 'ParentValues'] = vartable.at[vid, 'ParentValues']
            sum.at[vid, 'ChildValues']  = vartable.at[vid, 'ChildValues']

            clean = vartable.at[vid, 'Clean']
            if pd.notna(clean):
                sum.at[vid, 'Clean'] = list(clean.values())

            flagstr  = []
            cols     = dtable.columns(vid)
            colflags = {c : dtable.getFlags(c) for c in cols}
            flags    = set.union(*colflags.values())

            for flag in flags:
                if all([flag in colflags[c] for c in cols]):
                    flagstr.append(flag)
                else:
                    names = [c.name for c in cols if flag in colflags[c]]
                    flagstr.append('{} [{}]'.format(flag, ', '.join(names)))

            sum.at[vid, 'Flags'] = ';'.join(flagstr)

        sum.to_csv(args.summary_file, sep='\t')


def configLogging(args):
    """Configures ``funpack`` logging.

    :arg args: ``argparse.Namespace`` object containing parsed command line
               arguments.
    """

    # Custom log handler which
    # colours messages
    class LogHandler(logging.StreamHandler):

        def emit(self, record):

            levelno = record.levelno

            if   levelno >= logging.WARNING:  colour = '\x1b[31;1m'
            elif levelno >= logging.INFO:     colour = '\x1b[39;1m'
            elif levelno >= logging.DEBUG:    colour = '\x1b[90;1m'
            else:                             colour = ''

            # Reset terminal attributes
            # after each message.
            record.msg = '{}{}\x1b[0m'.format(colour, record.msg)

            return super(LogHandler, self).emit(record)

    logger = logging.getLogger('funpack')
    fmt    = logging.Formatter('%(asctime)s '
                               '%(levelname)8.8s '
                               '%(filename)20.20s '
                               '%(lineno)4d: '
                               '%(funcName)-15.15s - '
                               '%(message)s',
                               '%H:%M:%S')

    if args.log_file is None: handler = LogHandler()
    else:                     handler = logging.FileHandler(args.log_file)

    handler.setFormatter(fmt)
    logger.addHandler(handler)

    # configure verbosity
    if   args.quiet:      loglevel = logging.CRITICAL
    elif args.noisy == 0: loglevel = logging.INFO
    else:                 loglevel = logging.DEBUG

    logging.getLogger('funpack').setLevel(loglevel)

    if args.quiet or args.noisy < 3:
        warnings.filterwarnings('ignore',  module='pandas')
        warnings.filterwarnings('ignore',  module='numpy')
        warnings.filterwarnings('ignore',  module='tables')

    if args.noisy == 1:
        makequiet = ['funpack.expression',
                     'funpack.custom',
                     'funpack.cleaning_functions',
                     'funpack.processing_functions']
    elif args.noisy == 2:
        makequiet = ['funpack.expression',
                     'funpack.custom']
    else:
        makequiet = []

    for mod in makequiet:
        logging.getLogger(mod).setLevel(logging.INFO)


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
#
# util.py - Miscellaneous utility functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains a collection of miscellaneous utility functions and
constants.
"""


import re
import sys
import enum
import time
import shutil
import logging
import warnings
import functools
import contextlib
import subprocess      as sp
import multiprocessing as mp

try:
    import resource
except ImportError:
    resource = None

import numpy as np


log = logging.getLogger(__name__)


CTYPES = enum.Enum(
    'CTYPES',
    ['sequence',
     'integer',
     'continuous',
     'categorical_single',
     'categorical_single_non_numeric',
     'categorical_multiple',
     'categorical_multiple_non_numeric',
     'time',
     'date',
     'text',
     'compound',
     'unknown'])
"""The ``CTYPES`` enum defines all the types that ``funpack`` is aware of.
"""


DATA_TYPES = {

    # We have to use floating point for
    # integer types because pandas uses
    # nan to represent missing data.
    CTYPES.integer                          : np.float32,
    CTYPES.continuous                       : np.float32,
    CTYPES.categorical_single               : np.float32,
    CTYPES.categorical_multiple             : np.float32,
    CTYPES.sequence                         : np.uint32,
    CTYPES.categorical_single_non_numeric   : str,
    CTYPES.categorical_multiple_non_numeric : str,
    CTYPES.text                             : str,
    CTYPES.compound                         : str,
}
"""Default internal data type to use for the different variable types.
Used by the :func:`columnTypes` function. These types may be overridden
by the ``InternalType`` column of the variable table, which is populated
from the ``funpack/data/type.txt`` file (see :func:`.loadTableBases`).
"""


def parseColumnName(name):
    """Parses a UK Biobank column name, returns the components.

    Two column naming formats are supported. The name is expected to be
    a string of one of the following forms::

        variable-visit.instance
        variable.instance
        f.variable.visit.instance

    where ``variable`` and ``visit`` are integers. ``instance`` is typically
    also an integer, but non-numeric values for ``instance`` are
    accepted. This (and the second form above) is to allow parsing of derived
    columns (see e.g. the :func:`.binariseCategorical` processing function).

    Some variables have the form::

        f.variable..visit.instance

    For these variables, the visit is interpreted as a negative number.

    If ``name`` does not have one of the above forms, a :exc:`ValueError` is
    raised.

    .. note:: For the vast majority of biobank variables, the second number in
              a column name (``visit`` above) corresponds to the assessment
              visit. However, there are a small number of variables which are
              not associated with a specific visit, and thus for which this
              number does not correspond to a visit (e.g. variable 40006), but
              to some other coding.

              Confusingly, the UK Biobank showcase refers to the coding that a
              variable adheres to as an "instancing", whilst also using the
              term "instance" to refer to the columns of multi-valued
              variables - the ``instance`` element of the column name.

              The "instancing" that a variable uses is contained in the
              ``Instancing`` column of the variable table. Variables for which
              the ``visit`` component of their column names do correspond
              to an actual visit have an instancing equal to 2.

    :arg name: Column name
    :returns:  A tuple containing:
                - variable ID
                - visit number
                - instance (may be an integer or a string)
    """

    def parse_norm(grps):
        vid      = int(grps[0])
        visit    = int(grps[2])
        instance =     grps[3]
        if grps[1] is not None:
            visit = -visit
        return vid, visit, instance

    def parse_deriv(grps):
        vid      = int(grps[0])
        instance =     grps[1]
        return vid, 0, instance

    patterns = [
        (r'([0-9]+)-(-)?([0-9]+)\.(.+)',          parse_norm),
        (r'([0-9]+)\.(.+)',                       parse_deriv),
        (r'f\.([0-9]+)\.(\.)?([0-9]+)\.([0-9]+)', parse_norm)
    ]

    for pat, parse in patterns:

        pat   = re.compile(pat)
        match = pat.fullmatch(name)

        if match is None:
            continue

        vid, visit, instance = parse(match.groups())

        # accept numeric/non-numeric instance
        try:
            instance = int(instance)
        except ValueError:
            pass

        break

    if match is None:
        raise ValueError('Invalid column name: {}'.format(name))

    return (vid, visit, instance)


def generateColumnName(variable, visit, instance):
    """Generate a column name for the given variable, visit and instance.

    :arg variable: Integer variable ID
    :arg visit:    Visit number
    :arg instance: Instance number
    """
    return '{}-{}.{}'.format(variable, visit, instance)


def parseMatlabRange(r):
    """Parses a string containing a MATLAB-style ``start:stop`` or
    ``start:step:stop`` range, where the ``stop`` is inclusive).

    :arg r:   String containing MATLAB_style range.
    :returns: List of integers in the fully expanded range.
    """
    elems = [int(e) for e in r.split(':')]

    if len(elems) == 3:
        start, step, stop = elems
        if   step > 0: stop += 1
        elif step < 0: stop -= 1

    elif len(elems) == 2:
        start, stop  = elems
        stop        += 1
        step         = 1
    elif len(elems) == 1:
        start = elems[0]
        stop  = start + 1
        step  = 1
    else:
        raise ValueError('Invalid range string: {}'.format(r))

    return list(range(start, stop, step))


def dedup(seq):
    """Remove duplicates from a sequence, preserving order.
    Returns a list.
    """
    newseq = []
    for i in seq:
        if i not in newseq:
            newseq.append(i)
    return newseq


def wc(fname):
    """Uses ``wc`` to count the number of lines in ``fname``.

    :arg fname: Name of the file to check
    :returns:   Number of lines in ``fname``.
    """

    with timed('Row count', log):
        if shutil.which('wc'):
            nrows = sp.check_output(['wc', '-l', fname]).split()[0]
        else:
            nrows = 0
            with open(fname) as f:
                for _ in f:
                    nrows = nrows + 1
    return int(nrows)


def cat(files, outfile):
    """Uses ``cat`` to concatenate ``files``, saving the output to ``outfile``.

    :arg files:   Sequence of files to concatenate.
    :arg outfile: Name of file to save output to.
    """
    with timed('Concatenate files', log):
        if shutil.which('cat'):
            with open(outfile, 'w') as outf:
                cmd = ['cat'] + list(files)
                sp.run(cmd, check=True, stdout=outf)
        else:
            with open(outfile, 'w') as outf:
                for infile in files:
                    with open(infile, 'r') as inf:
                        while True:
                            buf = inf.read(16777216)
                            if buf:
                                outf.write(buf)
                            else:
                                break


def inMainProcess():
    """Returns ``True`` if the running process is the main (parent) process.
    Returns ``False`` if the running process is a child process (e.g. a
    ``multiprocessingg`` worker process).
    """
    return mp.current_process().pid == inMainProcess.pid


# Save the main process
# ID, so inMainProcess
# can compare against it
inMainProcess.pid = mp.current_process().pid



@contextlib.contextmanager
def timed(op=None, logger=None, lvl=None, fmt=None, minutes=None):
    """Context manager which times a section of code, and prints a log
    message afterwards.

    :arg op:      Name of operation which is being timed

    :arg logger:  Logger object to use - defaults to :attr:`log`.

    :arg lvl:     Log level - defaults to ``logging.INFO``.

    :arg fmt:     Custom message. If not provided, a default message is used.
                  Must be a ``'%'``-style format string which accepts two
                  parameters: the elapsed time (``%s``), and the memory usage
                  (``%i``)..

    :arg minutes: Deprecated, has no effect.
    """

    if minutes is not None:
        warnings.warn('The "minutes" argument is deprecated and has '
                      'no effect', DeprecationWarning, stacklevel=2)
    if fmt is None:
        fmt = '[{}] completed in %s (%+iMB)'.format(op)

    if logger is None:
        logger = log

    if lvl is None:
        lvl = logging.INFO

    if op is not None:
        logger.log(lvl, 'Running task [%s]', op)

    # ru_maxrss appears to be bytes under
    # macos, and kilobytes under linux
    if sys.platform == 'darwin': memdenom = 1048576.0
    else:                        memdenom = 1024.0

    if resource is not None:
        startmem  = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    else:
        startmem = 0

    starttime = time.time()

    yield

    endtime = time.time()

    hours   = int( (endtime - starttime) / 3600)
    minutes = int(((endtime - starttime) % 3600) / 60)
    seconds = int(((endtime - starttime) % 3600) % 60)
    timestr = '{:d} seconds'.format(seconds)

    if minutes > 0: timestr = '{} minutes, {}'.format(minutes, timestr)
    if hours   > 0: timestr = '{} hours, {}'  .format(hours,   timestr)

    if resource is not None:
        endmem  = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    else:
        endmem = 0

    mbytes = (endmem - startmem) / memdenom

    if minutes: logger.log(lvl, fmt, timestr, mbytes)
    else:       logger.log(lvl, fmt, timestr, mbytes)


def deprecated(message):
    """Decorator used to mark a function or method as deprecated """

    def wrapper(func):

        warnings.filterwarnings('default', category=DeprecationWarning)

        def decorator(*args, **kwargs):
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return functools.update_wrapper(decorator, func)

    return wrapper



#!/usr/bin/env python
#
# __init__.py - the importing package
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``importing`` package implements the data import stage - identifying
which columns are to be loaded, loading data from the input files, and
filtering/re-arranging the resulting data frame.

The :func:`.core.importData` function is the entry point to the import stage.
"""


from .core import   (importData,
                     NUM_ROWS,
                     MERGE_AXIS,
                     MERGE_STRATEGY,
                     MERGE_AXIS_OPTIONS,
                     MERGE_STRATEGY_OPTIONS)  # noqa
from .filter import (restrictVariables,)      # noqa



#!/usr/bin/env python
#
# demo.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import sys
import shutil
import tempfile
import os.path as op

import notebook.notebookapp as notebookapp
import bash_kernel.install  as bash_kernel_install


def main(argv=None):

    if argv is None:
        argv = sys.argv[1:]

    thisdir = op.abspath(op.dirname(__file__))
    nbdir = op.join(thisdir, 'demo')
    bash_kernel_install.main([])

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = op.join(tmpdir, 'demo')
        shutil.copytree(nbdir, tmpdir)
        notebookapp.main(['--notebook-dir', tmpdir] + argv)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# storage.py - The HDFStoreCollection class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`HDFStoreCollection` class, a simple
class which can be used to split the columns of a ``pandas.DataFrame``
across multiple HDF5 files.

.. note:: This module is not currently used.
"""


import multiprocessing       as mp
import multiprocessing.dummy as mpd
import functools             as ft
import itertools             as it
import os.path               as op
import                          uuid
import                          shutil
import                          logging
import                          tempfile
import                          warnings
import                          contextlib
import                          collections

import pandas  as pd
import numpy   as np


log = logging.getLogger(__name__)


COLUMNS_PER_FILE = 100
"""Default value for the ``colsPerFile`` parameter to
:class:`HDFStoreCollection`.
"""


@contextlib.contextmanager
def _silence_tables():
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', module='tables')
        yield


def is_bool_array(obj):
    ba = isinstance(obj, (np.ndarray, pd.Series))
    ba = ba and obj.dtype == np.bool
    return ba


class HDFStoreCollection(object):
    """The ``HDFStoreCollection`` class can be used to store, read and write
    one or more ``pandas.DataFrame`` objects (all with the same columns) as
    HDF5 files, where each file contains a sub-set of columns.

    If a ``multiprocessing.Manager`` is passed to :mod:`__init__`, the methods
    of the ``HDFStoreCollection`` are thread- and process-safe for reading and
    writing.
    """


    def __init__(self,
                 prefix=None,
                 workDir=None,
                 colsPerFile=None,
                 mgr=None):
        """Create a ``HDFStoreCollection``.


        :arg prefix:      Prefix to use for HDF files. Defaults to a randomly
                          generated sequence of characters.


        :arg workDir:     Directory to store the HDF files. Defaults to a
                          temporary directory, which will be deleted when this
                          ``HDFStoreCollection`` is garbage-collected.

        :arg colsPerFile: Maximum number of data columns to store in each
                          ``.h5`` file. Pytables doesn't allow more than
                          about 2000 columns to be stored in its HDF files,
                          so we have to split wide tables across multiple
                          files. Defaults to :attr:`COLUMNS_PER_FILE`.

        :arg mgr:         A ``multiprocessing.Manager`` - this must be provided
                          for concurrent write support. Not needed if
                          concurrent writing is not needed.
        """

        if prefix      is None: prefix      = uuid.uuid4().hex[:8]
        if colsPerFile is None: colsPerFile = COLUMNS_PER_FILE

        ownDir = workDir is None

        if ownDir:
            workDir = tempfile.mkdtemp()

        # __storeFiles is a list of file
        # names used to store the columns
        # of the data frame(s).
        #
        # __columns is a dict of
        # { column : file_index } mappings,
        # used to keep track of which
        # columns are stored in which file.
        #
        # __columns is a list of lists of
        # column names, used to keep track
        # of which columns are stored in
        # which file.
        #
        # __locks is a list of mgr.Lock
        # objects, used to lock files for
        # writing.
        #
        # All are initialised in the
        # append method.
        self.__mgr         = mgr
        self.__ownDir      = ownDir
        self.__prefix      = prefix
        self.__workDir     = op.abspath(workDir)
        self.__colsPerFile = colsPerFile
        self.__storeFiles  = None
        self.__fileLocks   = None
        self.__columns     = None

        # Ugly hack - columns which cannot be
        # stored in pytables (e.g. which have
        # a complicated data type) are stored
        # in-memory in the __unstorable dict.
        #
        # TODO If you need write support
        #      across multiple processes,
        #      you will need to use a more
        #      complicated (e.g. file-based)
        #      strategy for storing/sharing
        #      these columns.
        self.__unstorable = {}


    def __del__(self):
        """If this ``HDFStoreCollection`` created a temporary directory to
        store the HDF5 files, that directory is deleted.
        """
        if self.__ownDir:
            shutil.rmtree(self.__workDir)


    def __getstate__(self):
        """Returns the state of this ``HDFStoreCollection`` for pickling.
        """
        return (self.__prefix,
                self.__workDir,
                self.__colsPerFile,
                self.__storeFiles,
                self.__fileLocks,
                self.__columns,
                self.__unstorable)


    def __setstate__(self, state):
        """Set the state of this ``HDFStoreCollection`` for
        unpickling.
        """

        self.__ownDir      = False
        self.__mgr         = None
        self.__prefix      = state[0]
        self.__workDir     = state[1]
        self.__colsPerFile = state[2]
        self.__storeFiles  = state[3]
        self.__fileLocks   = state[4]
        self.__columns     = state[5]
        self.__unstorable  = state[6]


    @property
    def colsPerFile(self):
        """Returns the maximum number of columns stored in each HDF5 file. """
        return self.__colsPerFile


    @property
    def prefix(self):
        """Returns the file prefix used by this ``HDFStoreCollection``. """
        return self.__prefix


    @property
    def workDir(self):
        """Returns the directory that contains the HDF5 files. """
        return self.__workDir


    @contextlib.contextmanager
    def storeFile(self, idx, mode, *args, **kwargs):
        """Context manager which yields a ``pandas.HDFStore`` object
        for the file with the given ``idx``.
        """

        if self.__fileLocks is not None:
            lock = self.__fileLocks[idx]
        else:
            lock = mp.Lock()

        with lock:
            with pd.HDFStore(self.__storeFiles[idx],
                             mode=mode,
                             *args, **kwargs) as f:
                yield f


    def columnIndex(self, column):
        """Returns the index of the file that contains the given column. """
        for i in range(len(self.__columns)):
            if column in self.__columns[i]:
                return i
        raise ValueError('Column {} is not in store'.format(column))


    def columnBlocks(self, cols):
        """Splits the given sequence of column names into blocks for
        each underlying file.

        :returns: a ``dict`` of ``{ idx : [column] }`` mappings.
        """

        colblocks = collections.defaultdict(list)

        for col in cols:
            idx = self.columnIndex(col)
            colblocks[idx].append(col)

        return colblocks


    def columnsIn(self, idx):
        """Returns a list of all the columns stored in the specified file. """
        return self.__columns[idx]


    def _dropRows(self, rows, idx):
        """Drop the specified ``rows`` in the file at ``idx``. """

        log.debug('Dropping rows from store file %i: %s', idx, rows)

        cols = self.columnsIn(idx)
        with self.storeFile(idx, mode='a') as s:
            for col in cols:
                s.put(col, s[col][rows], format='table')


    def dropRows(self, rows, pool=None):
        """Drops the specified rows from the data set.

        :arg rows: Something which can be used to index the rows of a
                   ``pandas.Series``.
        """

        ownPool = pool is None

        if pool is None:
            pool = mpd.Pool(1)

        idxs = range(len(self.__storeFiles))
        pool.map(ft.partial(self._dropRows, rows), idxs)

        if ownPool:
            pool.close()
            pool.join()


    def _dropColumns(self, idx, cols):
        """Drop all columns from the specified file.

        Assumes that all columns in ``cols`` are present in the file at
        ``idx``.
        """

        log.debug('Dropping columns from store file %i: %s', idx, cols)

        with self.storeFile(idx, 'a') as s:
            for col in cols:
                s.remove(col)

                # self.__columns[idx] may be
                # a mp.Manager.list, so we
                # re-assign it to make sure
                # it is propagated.
                newcols = self.__columns[idx]
                newcols.remove(col)
                self.__columns[idx] = newcols


    def dropColumns(self, cols, pool=None):
        """Drops the specified columns from the data set.

        :arg cols: Sequence of columns to drop.
        """

        ownPool = pool is None

        if pool is None:
            pool = mpd.Pool(1)

        blocks = self.columnBlocks(cols)
        pool.starmap(self._dropColumns, blocks.items())

        if ownPool:
            pool.close()
            pool.join()


    def append(self, dataframe):
        """Append more rows to the store. """

        columns   = list(dataframe.columns)
        ncols     = len(columns)
        nfiles    = int(np.ceil(ncols / self.colsPerFile))
        colBlocks = []

        if self.__columns is not None:
            if self.columns != columns:
                raise ValueError('Columns do not match!')

        for i in range(nfiles):
            s = i * self.colsPerFile
            e = s + self.colsPerFile
            colBlocks.append(columns[s:e])

        if self.__storeFiles is None:

            filenames = ['{}_{:03d}.h5'.format(self.__prefix, i)
                         for i in range(nfiles)]
            filenames = [op.join(self.__workDir, f) for f in filenames]

            self.__storeFiles = filenames
            self.__columns    = collections.OrderedDict()

            if self.__mgr is not None:
                self.__fileLocks = [self.__mgr.Lock() for f in filenames]
                self.__columns   =  self.__mgr.list()
            else:
                self.__fileLocks = None
                self.__columns   = []

            self.__columns.extend([[] for i in range(nfiles)])

            for i, cb in enumerate(colBlocks):
                for col in cb:
                    # self.__columns[i] may be
                    # a mp.Manager.list, so we
                    # re-assign it to make sure
                    # it is propagated.
                    self.__columns[i] = self.__columns[i] + [col]

                log.debug('Assigning %u columns [%s, ..., %s] to file %s',
                          len(cb), cb[0], cb[-1], filenames[i])

        # TODO store index separately
        for i, block in enumerate(colBlocks):
            with self.storeFile(i, mode='a') as s, _silence_tables():

                log.debug('Appending %u columns [%s, ..., %s] to file %s',
                          len(block), block[0], block[-1], s.filename)

                for col in block:
                    s.append(col, dataframe[col], format='table')


    def __len__(self):
        """Returns the number of rows in the store. """
        return self.nrows


    @property
    def nrows(self):
        """Returns the number of rows in the store. """
        with self.storeFile(0, mode='r') as s:
            return s.get_storer(self.columns[0]).nrows


    @property
    def ncols(self):
        """Returns the number of columns in the store. """

        ncols = 0

        for i in range(len(self.__storeFiles)):
            with self.storeFile(i, mode='r') as s:
                ncols += len(s)

        return ncols


    @property
    def columns(self):
        """Returns all of the columns in the store. """
        return list(it.chain(*self.__columns))


    @property
    def index(self):
        """Returns the store index."""

        colblock = self.columnBlocks([self.columns[0]])
        idx, col = list(colblock.items())[0]

        with self.storeFile(idx, mode='r') as s:
            return s.select(col[0]).index


    @property
    def loc(self):
        """Emulates ``pandas.DataFrame.loc``. Read or write columns and
        rows from/to the store. See the :class:`StoreSlicer`.
        """
        return StoreSlicer(self, self.__unstorable)


class StoreSlicer(object):
    """The ``StoreSlicer`` provides a slice interface (``__getitem__`` and
    ``__setitem__``) to a :class:`HDFStoreCollection`.
    """

    def __init__(self, store, unstorable):
        """Create a ``StoreSlicer``.

        :arg store:      The :class:`HDFStoreCollection`.
        :arg unstorable: Dictionary which will be used to store columns
                         that cannot be stored on disk.
        """
        self.__store      = store
        self.__unstorable = unstorable


    def __normaliseSlice(self, slc):
        """Convert a slice object (as passed to :meth:`__getitem__` or
        :meth:`__setitem__`) into separate row/column indexers.
        """

        store = self.__store

        fancyTypes = (slice, list, tuple,
                      np.ndarray, pd.core.base.PandasObject)

        # some pandas indexing operations will
        # return a Series if a single column is
        # passed, or a DataFrame if list
        # containing a single column is passed.
        # So we need to differentiate between
        # these scenarios.
        onerow = False
        onecol = False

        # only rows, slice across all columns
        if not isinstance(slc, tuple):
            rows = slc
            cols = store.columns
        else:
            rows = slc[0]
            cols = slc[1]

            # single row index - make it a list
            if not (isinstance(rows, fancyTypes) or is_bool_array(rows)):
                onerow = True
                rows   = [rows]

            # column slice
            if isinstance(cols, slice):
                cols = pd.Index(store.columns)[cols]

            # single column label - make it a list
            elif not isinstance(cols, fancyTypes):
                onecol = True
                cols   = [cols]

        return rows, cols, onerow, onecol


    def __getitem__(self, slc):
        """Get the data at the given slice. """

        store                      = self.__store
        rows, cols, onerow, onecol = self.__normaliseSlice(slc)
        colblocks                  = store.columnBlocks(cols)
        results                    = []

        for idx, colblock in colblocks.items():
            with store.storeFile(idx, mode='r') as s:
                for col in colblock:
                    try:
                        series = s.select(col)
                    except Exception:
                        series = None

                    if series is None:
                        series = self.__unstorable.get(col, None)
                    results.append(series.loc[rows])

        result = pd.concat(results, axis=1, sort=False)

        if onecol: return result.iloc[:, 0]
        else:      return result
        return result


    def __setitem__(self, slc, val):
        """Set the data at the given slice. """

        store      = self.__store
        rows, cols = self.__normaliseSlice(slc)[:2]
        colblocks  = store.columnBlocks(cols)

        # If this assignment is across multiple columns,
        # does the provided value need to be indexed?
        if isinstance(val, pd.DataFrame):
            validxs = cols
        elif isinstance(val, np.ndarray) and val.shape == 2:
            validxs = range(len(cols))

        # constant/1D array/series - no indexing
        # necessary.
        else:
            validxs = [None] * len(cols)

        validxs = iter(validxs)

        for idx, colblock in colblocks.items():
            with store.storeFile(idx, mode='a') as s, _silence_tables():
                for col in colblock:

                    data   = s[col]
                    validx = next(validxs)
                    if validx is None: data[rows] = val
                    else:              data[rows] = val[validx]

                    try:
                        s.put(col, data, format='table')
                    except TypeError:
                        log.warning('Unstorable column: %s', col)
                        self.__unstorable[col] = data



#!/usr/bin/env python
#
# datatable.py - The DataTable class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DataTable` class, a container
class which holds a reference to the loaded data.
"""


import itertools             as it
import multiprocessing.dummy as mpd
import                          logging
import                          contextlib
import                          collections


from . import loadtables


log = logging.getLogger(__name__)


AUTO_VARIABLE_ID = 5000000
"""Starting variable ID to use for unknown data. Automatically generated
variable IDs really shoulld not conflict with actual UKB variable IDs.
"""


class Column:
    """The ``Column`` is a simple container class containing metadata
    about a single column in a data file.

    See the :func:`.parseColumnName` function for important information
    about column naming conventions in the UK BioBank.

    A fundamental assumption made throughout much of the ``funpack`` code-base
    is that columns with a variable ID (``vid``) equal to 0 are index columns
    (e.g. subject ID).

    A ``Column`` object has the following attributes:

     - ``datafile``: Input file that the column originates from (``None`` for
                     generated columns).

     - ``name``:     Column name as it appears in the input/output file(s)

     - ``index``:    Location of the column in the input/output file(s)

     - ``vid``:      Variable ID

     - ``visit``:    Visit number

     - ``instance``: Instance number/code

     - ``metadata``: Metadata which may be used to generate descriptive
                     information about the column

     - ``basevid``:  ID of the variable that the data in this column was
                     derived/generated from (and has the same data type as),
                     if not the original ``vid``.  For example, see the
                     ``take`` argument to :func:`.binariseCategorical`.

     - ``fillval``:  Fill value to use for missing values, instead of the
                     global default (e.g. as specified by
                     ``--tsv_missing_values``).
    """
    def __init__(self,
                 datafile,
                 name,
                 index,
                 vid=None,
                 visit=0,
                 instance=0,
                 metadata=None,
                 basevid=None,
                 fillval=None):

        if basevid is None:
            basevid = vid

        self.datafile = datafile
        self.name     = name
        self.index    = index
        self.vid      = vid
        self.visit    = visit
        self.instance = instance
        self.metadata = metadata
        self.basevid  = basevid
        self.fillval  = fillval


    def __str__(self):
        return 'Column({}, {}, {}, {}, {}, {})'.format(
            self.datafile,
            self.name,
            self.index,
            self.vid,
            self.visit,
            self.instance,
            self.basevid,
            self.fillval)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return (isinstance(other, Column)       and
                self.datafile == other.datafile and
                self.name     == other.name     and
                self.index    == other.index    and
                self.vid      == other.vid      and
                self.visit    == other.visit    and
                self.instance == other.instance and
                self.basevid  == other.basevid  and
                self.fillval  == other.fillval)



class DataTable:
    """The ``DataTable`` is a simple container class.

    It keeps references to the variable and processing tables, and the data
    itself. The :func:`importData` function creates and returns a
    ``DataTable``.

    A ``DataTable`` has the following attributes and helper methods:

    .. autosummary::
       :nosignature:

       pool
       manager
       parallel
       vartable
       proctable
       cattable
       columns
       allColumns
       present
       visits
       instances
       variables
       order


    Data should be accessed/modified via these methods:

    .. autosummary::
       :nosignature:

       index
       __getitem__
       __setitem__


    Columns can be added/removed, and rows removed, via these methods:

    .. autosummary::
       :nosignature:

       maskSubjects
       addColumns
       removeColumns

    Columns can be "flagged" with metadata labels via the :meth:`addFlag`
    method. All of the flags on a column can be retrieved via the
    :meth:`getFlags` method.


    The :meth:`subtable` method can be used to generate a replica
    ``DataTable`` with a specific subset of columns. It is intended for
    parallelisation, so that only the required data is copied over to tasks
    running in other processes. The :meth:`subtable` and :meth:`merge` methods
    are intended to be used like so:

      1. Create subtables which only contain data for specific columns::

             cols      = ['1-0.0', '2-0.0', '3-0.0']
             subtables = [dtable.subtable([c]) for c in cols]

      2. Use multiprocessing to perform parallel processing on each column::

             def mytask(dtable, col):
                 dtable[:, col] += 5
                 return dtable

             with dtable.pool() as pool:
                 subtables = pool.starmap(mytask, zip(subtables, cols))

      3. Merge the results back into the main table::

             for subtable in subtables:
                 dtable.merge(subtable)
    """


    def __init__(self,
                 data,
                 columns,
                 vartable,
                 proctable,
                 cattable,
                 pool=None,
                 mgr=None):
        """Create a ``DataTable``.

        :arg data:      ``pandas.DataFrame`` containing the data.
        :arg columns:   List of :class:`.Column` objects, representing the
                        columns that are in the data.
        :arg vartable:  ``pandas.DataFrame`` containing the variable table.
        :arg proctable: ``pandas.DataFrame`` containing the processing table.
        :arg cattable   ``pandas.DataFrame`` containing the category table.
        :arg pool:      ``multiprocessing.Pool`` for parallelising tasks.
        :arg mgr:       ``multiprocessing.Manager`` for sharing data between
                        tasks.
        """

        self.__data      = data
        self.__vartable  = vartable
        self.__proctable = proctable
        self.__cattable  = cattable
        self.__pool      = pool
        self.__mgr       = mgr
        self.__flags     = collections.defaultdict(set)

        # The varmap is a dictionary of
        # { vid : [Column] } mappings,
        # and the colmap is a dictionary
        # of { name : Column } mappings
        self.__varmap = collections.OrderedDict()
        self.__colmap = collections.OrderedDict()

        for col in columns:
            self.__colmap[col.name] = col
            if col.vid in self.__varmap: self.__varmap[col.vid].append(col)
            else:                        self.__varmap[col.vid] = [col]


    def __getstate__(self):
        """Returns the state of this :class:`.DataTable` for pickling. """
        return (self.__data,
                self.__vartable,
                self.__proctable,
                self.__cattable,
                self.__varmap,
                self.__colmap,
                self.__flags)


    def __setstate__(self, state):
        """Set the state of this :class:`.DataTable` for unpickling. """

        self.__data      = state[0]
        self.__vartable  = state[1]
        self.__proctable = state[2]
        self.__cattable  = state[3]
        self.__varmap    = state[4]
        self.__colmap    = state[5]
        self.__flags     = state[6]
        self.__pool      = None
        self.__mgr       = None


    @contextlib.contextmanager
    def pool(self):
        """Return a ``multiprocessing.Pool`` object for performing tasks
        in parallel on the data. For ``DataTable`` instances that have been
        unpickled (i.e. instances which are running in a sub-process), a
        ``multiprocessing.dummy.Pool`` instance is created and returned.
        """
        if self.__pool is not None:
            yield self.__pool
        else:
            with mpd.Pool(1) as pool:
                yield pool
                pool.close()
                pool.join()


    @property
    def manager(self):
        """Returns a ``multiprocessing.Manager`` for sharing data between
        processes.
        """
        if self.__mgr is None: return mpd.Manager()
        else:                  return self.__mgr


    @property
    def parallel(self):
        """Returns ``True`` if this invocation of ``funpack`` has the ability
        to run tasks in parallel, ``False`` otherwise.
        """
        return self.__pool is not None


    @property
    def vartable(self):
        """Returns the ``pandas.DataFrame`` containing the variable table. """
        return self.__vartable


    @property
    def proctable(self):
        """Returns the ``pandas.DataFrame`` containing the processing table.
        """
        return self.__proctable


    @property
    def cattable(self):
        """Returns the ``pandas.DataFrame`` containing the category table. """
        return self.__cattable


    @property
    def index(self):
        """Returns the subject indices."""
        return self.__data.index


    @property
    def shape(self):
        """Returns the (nrows, ncolumns) shape of the data."""
        return self.__data.shape


    @property
    def variables(self):
        """Returns a list of all integer variable IDs present in the data.

        The list includes the index variable (which has an id of ``0``).
        """
        return list(self.__varmap.keys())


    @property
    def allColumns(self):
        """Returns a list of all columns present in the data, including
        index columns.
        """
        return list(it.chain(*[self.columns(v) for v in self.variables]))


    @property
    def indexColumns(self):
        """Returns a list of all index columns present in the data. """
        return list(self.columns(0))


    @property
    def dataColumns(self):
        """Returns a list of all non-index columns present in the data. """
        return [c for c in self.allColumns if c.vid != 0]


    def order(self, vids):
        """Orders the data columns according to the given list of variable
        IDs.

        This method only affects the order in which columns and variables are
        returned from the :meth:`variables`, :meth:`allColumns`, and
        :meth:`dataColumns` methods - the order of the underlying dataframe
        is not changed.

        Variables which are in the data, but not in the ``vids`` list, are
        removed from the data set.

        :arg vids: Sequence of variable IDs in the desired order.
        """

        if not all([self.present(v) for v in vids]):
            raise ValueError('One of these variables is not '
                             'in the data: {}'.format(vids))

        oldvarmap = self.__varmap
        newvarmap = collections.OrderedDict()

        # index column stays
        newvarmap[0] = oldvarmap[0]

        # generate new variable map
        # (a dict of {vid : [Column]}
        # mappings)
        for vid in vids:
            newvarmap[vid] = self.columns(vid)

        # drop any columns associated
        # with variables that were not
        # listed in the new order.
        for vid, cols in list(oldvarmap.items()):
            if vid == 0:
                continue
            if vid not in vids:
                self.removeColumns(cols)

        self.__varmap = newvarmap


    def present(self, variable, visit=None, instance=None):
        """Returns ``True`` if the specified variable (and optionally visit/
        instance) is present in the data, ``False`` otherwise.
        """
        try:
            self.columns(variable, visit, instance)
            return True
        except KeyError:
            return False


    def columns(self, variable, visit=None, instance=None):
        """Return the data columns corresponding to the specified ``variable``,
        ``visit`` and ``instance``.

        :arg variable: Integer variable ID
        :arg visit:    Visit number. If ``None``, column names for all visits
                       are returned.
        :arg instance: Instance number. If ``None``, column names for all
                       instances are returned.
        :returns:      A list of :class:`.Column` objects.
        """

        cols = list(self.__varmap[variable])

        if visit is not None:
            cols = [c for c in cols if c.visit == visit]

        if instance is not None:
            cols = [c for c in cols if c.instance == instance]

        return cols


    def visits(self, variable):
        """Returns the visit IDs for the given ``variable``. """
        cols = self.columns(variable)
        return list({c.visit for c in cols})


    def instances(self, variable):
        """Returns the instance IDs for the given ``variable``. """
        cols = self.columns(variable)
        return list({c.instance for c in cols})


    def maskSubjects(self, mask):
        """Remove subjects where ``mask is False``. """
        self.__data = self.__data[mask]


    def removeColumns(self, cols):
        """Remove the columns described by ``cols``.

        :arg cols: Sequence of :class:`Column` objects to remove.
        """

        names = [c.name for c in cols]

        self.__data.drop(names, axis=1, inplace=True)

        for col in cols:

            vcols = self.__varmap[col.vid]
            vcols.remove(col)

            self.__colmap.pop(col.name)

            if len(vcols) > 0: self.__varmap[col.vid] = vcols
            else:              self.__varmap.pop(col.vid)


    def addColumns(self, series, vids=None, kwargs=None):
        """Adds one or more new columns to the data set.

        :arg series: Sequence of ``pandas.Series`` objects containing the
                     new column data to add.

        :arg vids:   Sequence of variables each new column is associated
                     with. If ``None`` (the default), variable IDs are
                     automatically assigned.

        :arg kwargs: Sequence of dicts, one for each series, containing
                     arguments to pass to the :class:`Column` constructor
                     (e.g. ``visit``, ``metadata``, etc).
        """

        if vids   is None: vids   = [None] * len(series)
        if kwargs is None: kwargs = [None] * len(series)

        for s in series:
            if s.name in self.__data.columns:
                raise ValueError(
                    'A column with name {} already exists - remove '
                    'it, or assign to it directly'.format(s.name))

        if len(vids) != len(series):
            raise ValueError('length of vids does not match series')

        startidx = len(self.__data.columns)
        idxs     = range(startidx, startidx + len(series))

        # if vids are not provided, auto-generate
        # a vid for each column starting from here.
        startvid = max(max(self.variables) + 1, AUTO_VARIABLE_ID)

        for s, idx, vid, kw in zip(series, idxs, vids, kwargs):

            if vid is None:
                vid      = startvid
                startvid = startvid + 1

            if kw is None:
                kw = {}

            col = Column(None, s.name, idx, vid, **kw)

            self.__data[  s.name] = s
            self.__colmap[s.name] = col

            # new column on existing variable.
            # We assume the data type is the
            # same as the existing columns for
            # this variable
            if vid in self.__varmap:
                self.__varmap[col.vid].append(col)

            # new variable - the addNewVariable
            # function will sort things out
            else:
                self.__varmap[col.vid] = [col]
                loadtables.addNewVariable(
                    self.__vartable, vid, col.name, s.dtype)


    def getFlags(self, col):
        """Return any flags associated with the specified column.

        :arg col: :class:`Column` object
        :returns: A ``set`` containing any flags associated with ``col``
        """
        return set(self.__flags[col])


    def addFlag(self, col, flag):
        """Adds a flag for the specified column.

        :arg col:  :class:`Column` object
        :arg flag: Flag to add
        """
        self.__flags[col].add(flag)


    def __getitem__(self, slc):
        """Get the specified slice from the data. This method has
        the same interface as the ``pandas.DataFrame.loc`` accessor.
        """
        return self.__data.loc[slc]


    def __setitem__(self, slc, value):
        """Set the specified slice in the data. This method has
        the same interface as the ``pandas.DataFrame.loc`` accessor.
        """
        self.__data.loc[slc] = value


    def __len__(self):
        """Returns the number of rows in the data set. """
        return len(self.__data)


    def subtable(self, columns=None, rows=None):
        """Return a new ``DataTable`` which only contains the data for
        the specified ``columns`` and/or ``rows``.

        This method can be used to create a replica ``DataTable`` where the
        underlying ``pandas.DataFrame`` only contains the specified
        columns. It is intended to be used when parallelising tasks, so that
        only necessary columns are copied back and forth between processes.

        :arg columns: List of :class:`Column` objects.
        :arg rows:    Sequence of row indices.
        :returns:     A new :class:`DataTable`.
        """

        if columns is None:
            columns  = self.dataColumns
            colnames = slice(None)
        else:
            colnames = [c.name for c in columns]

        columns = [self.allColumns[0]] + columns

        if rows is None:
            rows = slice(None)

        return DataTable(self.__data.loc[rows, colnames],
                         columns,
                         self.__vartable,
                         self.__proctable,
                         self.__cattable)


    def merge(self, subtable):
        """Merge the data from the given ``subtable`` into this ``DataTable``.

        :arg subtable: ``DataTable`` returned by :meth:`subtable`.
        """

        if self.shape[0] == subtable.shape[0]:
            subrows = slice(None)
        else:
            subrows = subtable.index

        if self.shape[1] == subtable.shape[1]:
            subcols = slice(None)
        else:
            subcols = [c.name for c in subtable.dataColumns]

        self[subrows, subcols] = subtable[subrows, subcols]

        for subcol in subtable.dataColumns:

            mycol               = self.__colmap[subcol.name]
            myflags             = self.__flags[mycol]
            subflags            = subtable.getFlags(subcol)
            self.__flags[mycol] = myflags.union(subflags)

            if subcol.metadata is not None:
                mycol.metadata = subcol.metadata



#!/usr/bin/env python
#
# test_dryrun.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path as op

import numpy  as np
import pandas as pd

import funpack.main as main

from . import gen_DataTable, gen_test_data, tempdir


def test_main_dryrun():
    with tempdir():
        gen_test_data(10, 100, 'data.tsv')
        assert main.main('--dry_run out.tsv data.tsv'.split()) == 0
        assert not op.exists('out.tsv')



#!/usr/bin/env python
#
# exporting_tsv.py - Export data to a TSV file.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :func:`exportTSV` function, which exports the data
contained in a :class:`.DataTable` to a TSV file.
"""


import functools as ft
import os.path   as op
import              os
import              logging

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

from . import util
from . import custom


log = logging.getLogger(__name__)


TSV_SEP = '\t'
"""Default separator string to use in TSV-style output files."""


NUM_ROWS = 10000
"""Default number of rows to export at a time by :func:`exportTSV` - the
default value for its ``numRows`` argument.
"""


@custom.exporter('tsv')
def exportTSV(dtable,
              outfile,
              idcol,
              colnames,
              sep=None,
              missingValues=None,
              dateFormat=None,
              timeFormat=None,
              formatters=None,
              numRows=None,
              nonNumericFile=None,
              **kwargs):
    """Export data to a TSV-style file.

    This may be parallelised by row - chunks of ``numRows`` rows will be
    saved to separate temporary output files in parallel, and then concatenated
    afterwards to produce the final output file.

    :arg dtable:         :class:`.DataTable` containing the data

    :arg outfile:        File to output to

    :arg idcol:          Name to use for the subject ID column

    :arg colnames:       Dict containing ``{oldcol : newcol}`` mappings

    :arg sep:            Separator character to use. Defaults to
                         :attr:`TSV_SEP`

    :arg missingValues:  String to use for missing/NA values. Defaults to the
                         empty string.

    :arg dateFormat:     Name of formatter to use for date columns.

    :arg timeFormat:     Name of formatter to use for time columns.

    :arg formatters:     Dict of ``{ [vid|column] : formatter }`` mappings,
                         specifying custom formatters to use for specific
                         variables.

    :arg numRows:        Number of rows to write at a time. Defaults to
                         :attr:`NUM_ROWS`.

    :arg nonNumericFile: If provided, non-numeric columns (after formatting)
                         are saved to this file instead of to ``outfile``
    """

    if sep           is None: sep           = TSV_SEP
    if missingValues is None: missingValues = ''
    if dateFormat    is None: dateFormat    = 'default'
    if timeFormat    is None: timeFormat    = 'default'
    if formatters    is None: formatters    = {}
    if numRows       is None: numRows       = NUM_ROWS

    # We're going to output each chunk of
    # subjects to a separate file (in
    # parallel), and then cat the files
    # together afterwards
    index      = dtable.index
    nchunks    = int(np.ceil(len(index) / numRows))
    idxchunks  = [index[i:i + numRows] for i in
                  range(0, len(index), numRows)]
    subtables  = [dtable.subtable(rows=c) for c in idxchunks]
    outfiles   = ['{}_{}'.format(outfile, i) for i in range(nchunks)]

    if nonNumericFile is not None:
        nnfiles = ['{}_{}'.format(nonNumericFile, i) for i in range(nchunks)]
    else:
        nnfiles = [None] * nchunks

    # write each chunk in parallel
    args = zip(subtables,
               outfiles,
               nnfiles,
               [True] + [False] * (nchunks - 1),
               range(nchunks))
    func = ft.partial(writeDataFrame,
                      idcol=idcol,
                      colnames=colnames,
                      sep=sep,
                      missingValues=missingValues,
                      dateFormat=dateFormat,
                      timeFormat=timeFormat,
                      formatters=formatters)

    with dtable.pool() as pool:
        pool.starmap(func, args)

    # concatenate the chunks to
    # form the final output file
    if len(outfiles) == 1:
        os.rename(outfiles[0], outfile)
    else:
        util.cat(outfiles, outfile)

    if nonNumericFile is not None and all([op.exists(f) for f in nnfiles]):
        if len(nnfiles) == 1:
            os.rename(nnfiles[0], nonNumericFile)
        else:
            util.cat(nnfiles, nonNumericFile)

    # remove intermediate files
    for f in outfiles + nnfiles:
        if f is not None and op.exists(f):
            os.remove(f)


def writeDataFrame(dtable,
                   outfile,
                   nonNumericFile,
                   header,
                   chunki,
                   idcol,
                   colnames,
                   sep,
                   missingValues,
                   dateFormat,
                   timeFormat,
                   formatters):
    """Writes all of the data in ``dtable`` to ``outfile``.

    Called by :func:`exportTSV` to output one chunk of data.

    :arg dtable:         :class:`.DataTable` containing the data

    :arg outfile:        File to output to

    :arg nonNumericFile: If provided, non-numeric columns (after formatting)
                         are saved to this file instead of to ``outfile``

    :arg header:         If ``True``, write the header row (column names).

    :arg idcol:          Name to use for the subject ID column

    :arg colnames:       Dict containing ``{oldcol : newcol}`` mappings

    :arg chunki:         Chunk index (used for logging)

    :arg sep:            Separator character to use. Defaults to
                         :attr:`TSV_SEP`

    :arg missingValues:  String to use for missing/NA values. Defaults to the
                         empty string.

    :arg dateFormat:     Name of formatter to use for date columns.

    :arg timeFormat:     Name of formatter to use for time columns.

    :arg formatters:     Dict of ``{ [vid|column] : formatter }`` mappings,
                         specifying custom formatters to use for specific
                         variables.
    """

    # If nonNumericFile is specified, we
    # store the names of all numeric and
    # non-numeric columns here so we can
    # figure out which columns to put
    # where. We run the columns through
    # formatting before deciding whether
    # they are numeric or non-numeric.
    numericCols    = []
    nonNumericCols = []

    columns = dtable.dataColumns
    towrite = pd.DataFrame(index=dtable.index)

    log.info('Writing %u columns and %u rows [chunk %u] to %s ...',
             len(columns), len(dtable), chunki, outfile)

    # Format every column, and
    # separate out into numeric
    # and non-numeric
    for col in columns:
        name   = colnames.get(col.name, col.name)
        series = formatColumn(
            col, dtable, dateFormat, timeFormat, formatters, chunki)

        towrite[name] = series

        # now we can tell whether this column
        # is numeric or non- numeric - store
        # its name accordingly
        if pdtypes.is_numeric_dtype(series): numericCols   .append(name)
        else:                                nonNumericCols.append(name)

    # Do we have any non-numeric columns?
    if (nonNumericFile is None) or (len(nonNumericCols) == 0):
        numericChunk   = towrite
        nonNumericFile = None

    else:
        numericChunk    = towrite[numericCols]
        nonNumericChunk = towrite[nonNumericCols]

        log.debug('Redirecting %i non-numeric columns to %s '
                  '(remaining %i columns will be written to %s)',
                  len(nonNumericCols), nonNumericFile,
                  len(numericCols),    outfile)

    if not header:
        idcol = None

    numericChunk.to_csv(outfile,
                        sep=sep,
                        na_rep=missingValues,
                        header=header,
                        index_label=idcol)

    if nonNumericFile is not None:
        nonNumericChunk.to_csv(nonNumericFile,
                               sep=sep,
                               na_rep=missingValues,
                               header=header,
                               index_label=idcol)


def formatColumn(col,
                 dtable,
                 dateFormat,
                 timeFormat,
                 formatters,
                 chunki):
    """Formats the data for the specified column.

    :arg col:        :class:`.Column` to format

    :arg dtable:     :class:`.DataTable` containing the data

    :arg dateFormat: Name of formatter to use for date columns.

    :arg timeFormat: Name of formatter to use for time columns.

    :arg formatters: Dict of ``{ [vid|column] : formatter }`` mappings,
                     specifying custom formatters to use for specific
                     variables.

    :arg chunki:     Output chunk index (used for logging).

    :returns:        ``pandas.Series`` instance containing the formatted data.
    """

    vid      = col.basevid
    vartable = dtable.vartable
    series   = dtable[:, col.name]

    # formatters can be specified
    # by VID or by column name
    formatter = formatters.get(vid, None)
    if formatter is None:
        formatter = formatters.get(col.name, None)

    if vid in vartable.index: vtype = vartable['Type'][vid]
    else:                     vtype = None

    # fall back to date/time formatting
    # if relevant for this column
    if formatter is None:
        if   vtype == util.CTYPES.date:
            formatter = dateFormat
        elif vtype == util.CTYPES.time or \
             pdtypes.is_datetime64_any_dtype(series):
            formatter = timeFormat

    if formatter is not None:
        log.debug('Formatting column %s [chunk %u] with %s formatter',
                  col.name, chunki, formatter)
        series = custom.runFormatter(formatter, dtable, col, series)

    # apply column-specific fill
    # value, if there is one
    if col.fillval is not None:
        series.fillna(value=col.fillval, inplace=True)

    return series



#!/usr/bin/env python
#
# processing_functions.py - Processing functions
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains definitions of processing functions - functions which
may be specifeid in the processing table.


A processing function may perform any sort of processing on one or more
variables. A processing function may add, remove, or manipulate the columns of
the :class:`DataTable`.


All processing functions must accept the following as their first two
positional arguments:


 - The :class:`.DataTable` object, containing references to the data, variable,
   and processing table.
 - A list of integer ID of the variables to process.


Furthermore, all processing functions must return one of the following:

 - ``None``, indicating that no columns are to be added or removed.

 - A ``list`` (must be a ``list``) of :class:`.Column` objects describing the
   columns that should be removed from the data.

 - A ``tuple`` (must be a ``tuple``) of length 2, containing:

    - A list of ``pandas.Series`` that should be added to the data.

    - A list of variable IDs to use for each new ``Series``. This list must
      have the same length as the list of new ``Series``, but if they are not
      associated with any specific variable, ``None`` may be used.

 - A ``tuple`` of length 3, containing:

    - List of columns to be removed
    - List of ``Series`` to be added
    - List of variable IDs for each new ``Series``.

 - A ``tuple`` of length 4, containing the above, and:

    - List of dicts associated with each of the new ``Series``. These will be
      passed as keyword arguments to the :class:`.Column` objects that
      represent each of the new ``Series``.

The following processing functions are defined:

 .. autosummary::
   :nosignatures:

   removeIfSparse
   removeIfRedundant
   binariseCategorical
   expandCompound
"""


import multiprocessing as mp
import functools       as ft
import itertools       as it
import                    logging
import                    collections

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

from . import processing_functions_core as core
from . import                              util
from . import                              custom


log = logging.getLogger(__name__)


@custom.processor()
def removeIfSparse(dtable,
                   vids,
                   minpres=None,
                   minstd=None,
                   mincat=None,
                   maxcat=None,
                   abspres=True,
                   abscat=True,
                   ignoreType=False):
    """removeIfSparse([minpres], [minstd], [mincat], [maxcat], [abspres], [abscat])
    Removes columns deemed to be sparse.

    Removes columns for the variables in ``vids`` if they are sparse.

    :arg ignoreType: Defaults to ``False``. If ``True``, all specified tests are
                     run regardless of the types of the ``vids``.

    See the :func:`.isSparse` function for details on the other arguments.
    """  # noqa

    remove = []

    for vid in vids:

        if ignoreType: vtype = None
        else:          vtype = dtable.vartable.loc[vid, 'Type']

        for col in dtable.columns(vid):

            log.debug('Checking column %s for sparsity', col.name)

            isSparse, test, val = core.isSparse(dtable[:, col.name],
                                                vtype,
                                                minpres=minpres,
                                                minstd=minstd,
                                                mincat=mincat,
                                                maxcat=maxcat,
                                                abspres=abspres,
                                                abscat=abscat)

            if isSparse:
                log.debug('Dropping sparse column %s (%s: %f)',
                          col.name, test, val)
                remove.append(col)

    if len(remove) > 0:
        log.debug('Dropping %u sparse columns: %s ...',
                  len(remove), remove[:5])

    return remove


@custom.processor()
def removeIfRedundant(dtable, vids, corrthres, nathres=None):
    """removeIfRedundant(corrthres, [nathres])
    Removes columns deemed to be redundant.

    Removes columns from the variables in ``vids`` if they are redundant.

    Redundancy is determined by calculating the correlation between pairs
    of columns - see the :func:`.isRedundant` function.
    """

    # Ignore non-numeric columns
    cols     = list(it.chain(*[dtable.columns(v) for v in vids]))
    cols     = [c for c in cols if pdtypes.is_numeric_dtype(dtable[:, c.name])]
    colnames = [c.name for c in cols]

    # evaluate all pairs at once
    if not dtable.parallel:
        colpairs = list(it.combinations(colnames, 2))
        log.debug('Checking %u columns for redundancy', len(cols))
        redundant = core.redundantColumns(
            dtable[:, :], colpairs, corrthres=corrthres, nathres=nathres)

    # evaluate in parallel
    else:
        # We need to compare every pair of
        # columns. We parallelise this by
        # splitting the columns into chunks,
        # and then passing pairs of chunks
        # to the redundantColumns function.
        ncols     = len(cols)
        chunksize = max(2, int(np.ceil(ncols / mp.cpu_count())))
        chunks    = [colnames[i:i + chunksize]
                     for i in range(0, ncols, chunksize)]

        # we build a list of column
        # pairs, and views into the
        # dataframe, for each chunk
        chunkpairs = []
        chunkdata  = []

        # while doing so, we keep track
        # of column pairs that have been
        # assigned to a chunk, so we don't
        # evaluate any pair more than once
        assignedcolpairs = set()

        # for every chunk, and
        # every pair of chunks
        for chunk1, chunk2 in it.combinations_with_replacement(chunks, 2):
            chunkcols   = util.dedup(chunk1 + chunk2)
            ichunkpairs = []
            for colpair in it.combinations(chunkcols, 2):
                if colpair not in assignedcolpairs:
                    ichunkpairs     .append(colpair)
                    assignedcolpairs.add(   colpair)
            if len(ichunkpairs) > 0:
                ichunkcols = util.dedup(it.chain(*ichunkpairs))
                ichunkdata = dtable[:, ichunkcols]
                chunkpairs.append(ichunkpairs)
                chunkdata .append(ichunkdata)

        log.debug('Checking %u columns for redundancy (%u tasks)',
                  len(cols), len(chunkdata))

        func = ft.partial(core.redundantColumns,
                          corrthres=corrthres, nathres=nathres)
        with dtable.pool() as pool:
            redundant = pool.starmap(func, zip(chunkdata, chunkpairs))
            redundant = list(it.chain(*redundant))

    redundant = util.dedup(sorted(redundant))

    if len(redundant) > 0:
        log.debug('Dropping %u redundant columns: %s ...',
                  len(redundant), redundant[:5])

    cols = collections.OrderedDict(zip(colnames, cols))
    return [cols[r] for r in redundant]


@custom.processor(auxvids=['take'])
def binariseCategorical(dtable,
                        vids,
                        acrossVisits=False,
                        acrossInstances=True,
                        minpres=None,
                        nameFormat=None,
                        replace=True,
                        take=None,
                        fillval=None,
                        replaceTake=True):
    """binariseCategorical([acrossVisits], [acrossInstances], [minpres], [nameFormat], [replace])
    Replace a categorical column with one binary column per category.

    Binarises categorical variables - replaces their columns with
    one new column for each value, containing ``1`` for subjects
    with that value, and ``0`` otherwise.

    :arg dtable:          The :class:`.DataTable`

    :arg vids:            Sequence of variable IDs to (independently) apply the
                          binarisation to.

    :arg acrossVisits:    If ``True``, the binarisation is applied across
                          visits for each variable.

    :arg acrossInstances: If ``True``, the binarisation is applied across
                          instances for each variable.

    :arg minpres:         Optional threshold - categorical values with less
                          than this many occurrences will not be added as
                          columns.

    :arg nameFormat:      Format string defining how the new columns should
                          be named - see below.

    :arg replace:         If ``True`` (the default), the original columns are
                          returned for removal.

    :arg take:            Optional variable ID, or sequence of variable IDs
                          (one for each of the main ``vids``) to take values
                          from. If provided, the generated columns will have
                          values from the column(s) of this variable, instead
                          of containinng binary 0/1 values. A ``take``
                          variable must have columns that match the columns of
                          the corresponding ``vid`` (by visits and instances).

    :arg fillval:         If ``take`` is provided, the value to use for
                          ``False`` rows. Defaults to ``np.nan``

    :arg replaceTake:     If ``True`` (the default), and ``takeFrom`` variables
                          were specified, the columns associated with the
                          ``take`` variables are returned for removal.

    The ``nameFormat`` argument controls how the new data columns should be
    named - it must be a format string using named replacement fields
    ``'vid'``, ``'visit'``, ``'instance'``, and ``'value'``. The ``'visit'``
    and ``'instance'`` fields may or may not be necessary, depending on the
    value of the ``acrossVisits`` and ``acrossInstances`` arguments.

    The default value for the ``nameFormat`` string is as follows:

    ================ =================== ======================================
    ``acrossVisits`` ``acrossInstances`` ``nameFormat``
    ================ =================== ======================================
    ``False``        ``False``           ``'{vid}-{visit}.{instance}_{value}'``
    ``False``        ``True``            ``'{vid}-{visit}.{value}'``
    ``True``         ``False``           ``'{vid}-{value}.{instance}'``
    ``True``         ``True``            ``'{vid}-0.{value}'``
    ================ =================== ======================================
    """  # noqa

    # get groups of columns for vid, grouped
    # according to acrossVisits/acrossInstances
    def gatherColumnGroups(vid):
        colgroups = []
        visits    = dtable.visits(   vid)
        instances = dtable.instances(vid)
        if not (acrossVisits or acrossInstances):
            for visit, instance in it.product(visits, instances):
                colgroups.append(dtable.columns(vid, visit, instance))
        elif acrossInstances and (not acrossVisits):
            for visit in visits:
                colgroups.append(dtable.columns(vid, visit))
        elif (not acrossInstances) and acrossVisits:
            for instance in instances:
                colgroups.append(dtable.columns(vid, instance=instance))
        else:
            colgroups = [dtable.columns(vid)]
        return colgroups

    defaultNameFormat = {
        (False, False) : '{vid}-{visit}.{instance}_{value}',
        (False, True)  : '{vid}-{visit}.{value}',
        (True,  False) : '{vid}-{value}.{instance}',
        (True,  True)  : '{vid}-0.{value}',
    }

    if nameFormat is None:
        nameFormat = defaultNameFormat[acrossVisits, acrossInstances]

    if not isinstance(take, collections.Sequence):
        take = [take] * len(vids)

    if len(take) != len(vids):
        raise ValueError('take must be either None, a single variable ID, '
                         'or a list of variable IDs, one for each of the '
                         'main vids.')

    remove     = []
    newseries  = []
    newvids    = []
    newcolargs = []

    for vid, takevid in zip(vids, take):

        colgrps = gatherColumnGroups(vid)

        if takevid is None: takegrps = [None] * len(colgrps)
        else:               takegrps = gatherColumnGroups(takevid)

        for cols, takecols in zip(colgrps, takegrps):

            if takecols is None: tkdata = None
            else:                tkdata = dtable[:, [c.name for c in takecols]]

            data              = dtable[:, [c.name for c in cols]]
            binarised, values = core.binariseCategorical(data,
                                                         minpres=minpres,
                                                         take=tkdata)

            if replace:                                remove.extend(cols)
            if replaceTake and (takecols is not None): remove.extend(takecols)

            for col, val in zip(binarised.T, values):

                # make sure no periods appear
                # in the resulting column name.
                # We're assuming here that all
                # categoricals are integers,
                # which has not been verified.
                try:               val = int(val)
                except ValueError: pass

                fmtargs = {
                    'vid'      : str(int(cols[0].vid)),
                    'visit'    : str(int(cols[0].visit)),
                    'instance' : str(int(cols[0].instance)),
                    'value'    : str(val)
                }

                series = pd.Series(
                    col,
                    index=dtable.index,
                    name=nameFormat.format(**fmtargs))

                colargs = {
                    'metadata' : val,
                    'basevid'  : takevid,
                    'fillval'  : fillval
                }

                newvids   .append(vid)
                newcolargs.append(colargs)
                newseries .append(series)

    return remove, newseries, newvids, newcolargs


@custom.processor()
def expandCompound(dtable, vids, nameFormat=None, replace=True):
    """expandCompound([nameFormat], [replace])
    Expand a compound column into a set of columns, one for each value.

    Expands compound variables into a set of columns, one for each value.
    Rows with different number of values are padded with ``np.nan``.

    :arg dtable:     The :class:`.DataTable`

    :arg vids:       Sequence of variable IDs to (independently) apply the
                     expansion to.

    :arg nameFormat: Format string defining how the new columns should be named
                     - see below.

    :arg replace:    If ``True`` (the default), the original columns are
                     flagged for removal.
    """

    if nameFormat is None:
        nameFormat = '{vid}-{visit}.{instance}_{index}'

    columns   = list(it.chain(*[dtable.columns(v) for v in vids]))
    newseries = []
    newvids   = []

    for column in columns:

        data    = dtable[:, column.name]
        newdata = core.expandCompound(data)

        for i in range(newdata.shape[1]):

            coldata = newdata[:, i]
            name    = nameFormat.format(vid=column.vid,
                                        visit=column.visit,
                                        instance=column.instance,
                                        index=i)

            newvids  .append(column.vid)
            newseries.append(pd.Series(coldata,
                                       index=dtable.index,
                                       name=name))

    if replace: return columns, newseries, newvids
    else:       return          newseries, newvids



#!/usr/bin/env python
#
# cleaning_functions.py - Cleaning functions.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains definitions of cleaning functions - functions
which may be specified in the ``Clean`` column of the variable table,
and which are appiled during the data import stage.

The following cleaning functions are available:

.. autosummary::
   :nosignatures:

   remove
   keepVisits
   fillVisits
   fillMissing
   convertICD10Codes
   makeNa


All cleaning functions (with two exceptions - :func:`remove` and
:func:`keepVisits`, explained below) will be passed the following as their
first positional arguments, followed by any arguments specified in the
variable table:

 - The :class:`.DataTable` object, containing references to the data, variable,
   and processing table.
 - The integer ID of the variable to process

These functions are expected to perform in-place pre-processing on the data.


The :func:`remove` and :func:`keepVisits` functions are called before the data
is loaded, as they are used to determine which columns should be loaded
in. They therefore have a different function signature to the other cleaning
functions, which are called after the data has been loaded. These functions
are passed the following positional arguments, followed by arguments specified
in the variable table:

 - The variable table, containing metadata about all known variables

 - The integer ID of the variable to be checked

 - A list of :class:`.Column` objects denoting all columns
   associated with this variable that are present in the data

These functions are expected to return a modified list containing the columns
that should be loaded.
"""


import logging

import numpy            as np
import pandas           as pd
import pandas.api.types as pdtypes

from . import util
from . import icd10
from . import custom
from . import hierarchy
from . import expression


log = logging.getLogger(__name__)


@custom.cleaner()
def remove(vartable, vid, columns):
    """remove()
    Remove (do not load) all columns associated with ``vid``.

    :arg vartable: Variable table
    :arg vid:      Integer variable ID.
    :arg columns:  List of all :class:`.Column` objects associated with
                   ``vid``.
    :returns:      An empty list.
    """
    return []


@custom.cleaner()
def keepVisits(vartable, vid, columns, *tokeep):
    """keepVisits(visit, [visit, ...])
    Only load columns for ``vid`` which correspond to the specified visits.

    This test is only applied to variables which have an *instancing* equal to
    2:

    https://biobank.ctsu.ox.ac.uk/crystal/instance.cgi?id=2

    Such variables are not associated with a specific visit, so it makes no
    sense to apply this test to them. See the
    :func:`.loadtables.addNewVariable` function for more information on
    variable instancing.

    :arg vartable: Variable table

    :arg vid:      Integer variable ID.

    :arg columns:  List of all :class:`.Column` objects that are associated
                   with ``vid``.

    :arg tokeep:   Visit IDs (integers), or ``'first'``, or ``'last'``,
                   indicating that the first or last visit should be loaded.

    :returns:      List of columns that should be loaded.
    """

    # variables which don't follow instancing
    # 2 do not have columns that correspond
    # to specific visits.
    instancing = vartable.loc[vid, 'Instancing']
    if instancing != 2:
        return columns

    keep      = []
    minVisit  = min([c.visit for c in columns])
    maxVisit  = max([c.visit for c in columns])

    for col in columns:

        test = [col.visit]

        if col.visit == minVisit: test.append('first')
        if col.visit == maxVisit: test.append('last')

        if col not in keep and any([v in tokeep for v in test]):
            keep.append(col)

    return keep


@custom.cleaner()
def fillVisits(dtable, vid, method='mode'):
    """fillVisits([method=(mode|mean)])
    Fill missing visits from available visits.

    For a variable with multiple visits, fills missing values from the visits
    that do contain data.
    """

    if method not in ('mode', 'mean'):
        raise ValueError('Unknown method: {}'.format(method))

    instances = dtable.instances(vid)

    for instance in instances:
        columns = [c.name for c in dtable.columns(vid, instance=instance)]
        view    = dtable[:, columns]

        if   method == 'mode': repvals = view.mode('columns')[0]
        elif method == 'mean': repvals = view.mean('columns')

        log.debug('Filling NA values in columns %s across visits', columns)

        for col in columns:
            dtable[:, col] = dtable[:, col].fillna(repvals)


@custom.cleaner()
def fillMissing(dtable, vid, value):
    """fillMissing(fill_value)
    Fill missing values with a constant.

    Fills all missing values in columns for the given variable with
    ``value``.
    """
    columns = [c.name for c in dtable.columns(vid)]
    log.debug('Filling NA values in columns %s with %i', columns, value)
    for col in columns:
        dtable[:, col] = dtable[:, col].fillna(value)


@custom.cleaner()
def flattenHierarchical(dtable, vid, level=None, name=None):
    """flattenHierarchical([level], [name])
    Replace leaf values with parent values in hierarchical variables.

    For hierarchical variables such as the ICD10 disease categorisations,
    this function replaces leaf values with a parent value.

    The ``name`` argument allows the hierarchy data type to be specified -
    see the :attr:`.HIERARCHY_DATA_NAMES` dictionary.

    The ``level`` argument allows the depth of the parent value to be selected
    - ``0`` (the default) replaces a value with the top-level (i.e. its most
    distant) parent value, ``1`` the second-level parent value, etc.

    If, for a particular value, the ``level`` argument is greater than the
    number of ancestors of the value, the value is unchanged.
    """

    if level is None:
        level = 0

    columns = [c.name for c in dtable.columns(vid)]
    vhier   = hierarchy.getHierarchyFilePath(dtable, vid, name)
    vhier   = hierarchy.loadHierarchyFile(vhier)

    def index_parent(parents, coding):

        # the list returned by Hierarchy.parents
        # is ordered from closest to most distant
        # parent, so we index from the end.
        if pdtypes.is_number(parents) and pd.isna(parents):
            return coding

        idx = len(parents) - 1 - level

        # If the specified level is out of
        # bounds, we return the original value
        if idx < 0: return coding
        else:       return parents[idx]

    log.debug('Flattening hierarchical data in columns %s (level %i)',
              columns, level)
    for col in columns:
        data           = dtable[:, col]
        parents        = data.map(vhier.parents, na_action='ignore')
        parents        = parents.combine(data, index_parent)
        dtable[:, col] = parents


@custom.cleaner()
def codeToNumeric(dtable, vid, name=None):
    """codeToNumeric()
    Convert hierarchical alpha-numeric codes to numeric equivalents.

    Given a hierarchical variable which contains alpha-numeric codings, this
    function will replace the codings with numeric equivalents.

    :arg name: Data coding name (see :attr:`.HIERARCHY_DATA_NAMES`).
    """

    log.debug('Converting hierarchy codes to numeric for variable %i', vid)

    hier = hierarchy.getHierarchyFilePath(dtable, vid, name=name)
    hier = hierarchy.loadHierarchyFile(hier)

    def convert(code):
        return hierarchy.codeToNumeric(
            code, dtable=dtable, vid=vid, name=name)

    cols  = [c.name for c in dtable.columns(vid)]
    codes = dtable[:, cols]

    for col in cols:
        dtable[:, col] = codes[col].map(convert)

    if name is None:
        coding = dtable.vartable.loc[vid, 'DataCoding']
        ncmap  = {c : n for n, c in hierarchy.HIERARCHY_DATA_NAMES.items()}
        name   = ncmap.get(coding, None)

    if name == 'icd10':
        for col in cols:
            icd10.storeCodes(codes[col].tolist())


@custom.cleaner()
@util.deprecated('Use codeToNumeric() instead')
def convertICD10Codes(dtable, vid):
    """convertICD10Codes()
    Convert ICD10 codes to numeric values. Deprecated - use codeToNumeric.

    Converts values which are assumed to contain ICD10 codes into numeric
    values.
    """
    codeToNumeric(dtable, vid, 'icd10')


@custom.cleaner()
def parseSpirometryData(dtable, vid):
    """parseSpirometryData()
    Parse spirometry (lung volume measurement) data.

    Parses values which contain spirometry (lung volume measurement) test
    data.

    Columns for UK Biobank variable 3066 and 10697 contain comma-separated
    time series data. This function converts the text values into ``numpy``
    arrays, and potentially re-orders instances of the variable within each
    visit - subjects potentially underwent a series of tests, with each test
    stored as a separate instance within one visit. However, the instance
    order does not necessarily correspond to the test order.

    See the following links for more information:

     - https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=3066
     - https://biobank.ctsu.ox.ac.uk/crystal/field.cgi?id=10697
    """
    visits = dtable.visits(vid)

    # applied to each row
    def fixOrder(vals):

        def parseBlow(val):
            # the first entry is "blowN",
            # where N is the blow number
            try:
                tkn = val[:10].split(',', maxsplit=1)[0]
                return int(tkn[4:])
            except Exception:
                return np.nan
        bnumvals = [(parseBlow(v), v) for v in vals]
        bnumvals = sorted(bnumvals)
        return pd.Series([bv[1] for bv in bnumvals],
                          index=vals.index)

    def parseValue(val):
        # first entry is "blowN", second
        # entry is number of samples.
        # Ignore them both, turn the
        # remainder into a numpy array
        try:
            val = val.split(',', maxsplit=2)[2]
            return np.fromstring(val, sep=',')
        except Exception:
            return np.nan

    for visit in visits:
        cols = dtable.columns(vid, visit=visit)
        cols = [c.name for c in cols]

        data = dtable[:, cols]
        data = data.apply(fixOrder,   axis=1)
        for col in cols:
            data.loc[:, col] = data.loc[:, col].apply(parseValue)

        dtable[:, cols] = data


@custom.cleaner()
def makeNa(dtable, vid, expr):
    """makeNa(expression)
    Replace values which pass the expression with NA.

    Replaces all values in ``vid`` for which ``func`` evaluates to ``True``
    with ``nan``.
    """

    expr = expression.Expression('v{} {}'.format(vid, expr))
    cols = dtable.columns(vid)

    for col in cols:
        col               = col.name
        mask              = expr.evaluate(dtable[:], {vid : col})
        dtable[mask, col] = np.nan



#!/usr/bin/env python
#
# test_main_import_all.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import textwrap  as tw

import numpy as np

import pandas as pd

import funpack.main as main

from . import (patch_logging,
               tempdir,
               gen_test_data)


@patch_logging
def test_main_import_all():

    with tempdir():

        gen_test_data(10, 20, 'data.tsv')

        main.main('-nb out.tsv data.tsv'.split())
        baseline = pd.read_csv('out.tsv', delimiter='\t', index_col=0)

        main.main('--import_all -nb -ow out.tsv data.tsv'.split())
        passthru = pd.read_csv('out.tsv', delimiter='\t', index_col=0)
        assert np.all(baseline == passthru)

        main.main('-v 1:3 --import_all -nb -ow out.tsv data.tsv'.split())
        dropvars = pd.read_csv('out.tsv', delimiter='\t', index_col=0)
        exp = baseline.loc[:, ['1-0.0', '2-0.0', '3-0.0']]
        assert np.all(dropvars == exp)

        main.main('-v 3:-1:1 --import_all -nb -ow out.tsv data.tsv'.split())
        ordered = pd.read_csv('out.tsv', delimiter='\t', index_col=0)
        exp = baseline.loc[:, ['3-0.0', '2-0.0', '1-0.0']]
        assert np.all(ordered == exp)


        vartable = tw.dedent("""
        ID\tType\tDescription\tDataCoding\tNAValues\tRawLevels\tNewLevels\tParentValues\tChildValues\tClean
        3
        4
        5
        """).strip()

        with open('variables.tsv', 'wt') as f:
            f.write(vartable)

        main.main('-r -nb -vf variables.tsv --import_all -ow out.tsv data.tsv'.split())
        ordered = pd.read_csv('out.tsv', delimiter='\t', index_col=0)
        exp = baseline.loc[:, ['3-0.0', '4-0.0', '5-0.0']]
        assert np.all(ordered == exp)


@patch_logging
def test_main_import_all_variable_removed():

    data = tw.dedent("""
    f.eid\t1-0.0\t2-0.0
    1\t5\t10
    2\t6\t
    3\t7\t
    4\t8\t
    5\t9\t
    6\t10\t
    7\t11\t
    8\t12\t
    9\t13\t
    10\t14\t
    """).strip()

    with tempdir():
        with open('data.tsv', 'wt') as f:
            f.write(data)

        main.main('-ia -v 1:2 -sp -apr all removeIfSparse(minpres=5) '
                  'out.tsv data.tsv'.split())

        got = pd.read_csv('out.tsv', delimiter='\t', index_col=0)
        assert np.all(got.index == np.arange(1, 11))
        assert got.columns == ['1-0.0']
        assert np.all(got['1-0.0'] == np.arange(5, 15))



@patch_logging
def test_main_importAll_column_patterns():

    data = tw.dedent("""
    eid,col1,column2,col3
    1,11,21,31
    2,12,22,32
    3,13,23,33
    4,14,24,34
    5,15,25,35
    """).strip()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        main.main('-nb -ia -co column2 -co col3 out1.txt data.txt'.split())
        main.main('-nb -ia -co col?             out2.txt data.txt'.split())

        got1 = pd.read_csv('out1.txt', delimiter='\t', index_col=0)
        got2 = pd.read_csv('out2.txt', delimiter='\t', index_col=0)

        assert np.all(got1.columns == ['column2', 'col3'])
        assert np.all(got2.columns == ['col1',    'col3'])


def test_main_import_all_variables_column_patterns():

    data = tw.dedent("""
    eid,1-0.0,2-0.0,3-0.0,col4
    1,11,21,31,41
    2,12,22,32,42
    3,13,23,33,43
    4,14,24,34,44
    5,15,25,35,45
    """).strip()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        main.main('-nb -ia -v 3 -v 2 -v 1 -co col4 out.txt data.txt'.split())

        exp = pd.read_csv('data.txt', delimiter=',',  index_col=0)
        got = pd.read_csv('out.txt',  delimiter='\t', index_col=0)

        with open('out.txt') as f:
            hdr = f.readline().strip()

        assert hdr == '\t'.join(['eid', 'col4', '3-0.0', '2-0.0', '1-0.0'])

        exp = exp[['col4', '3-0.0', '2-0.0', '1-0.0']]
        assert (exp == got).all().all()



#!/usr/bin/env python
#
# filter.py - functions for filtering columns/rows.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
"""This module contains functions used by the :func:`.core.importData`
function to identify which columns should be imported, and to filter
rows from a data frame after it has been loaded.
"""


import functools as ft
import itertools as it
import              fnmatch
import              logging
import              collections

from .. import      expression
from .. import      loadtables
from .. import      fileinfo


log = logging.getLogger(__name__)


def restrictVariables(cattable, variables, categories):
    """Determines which variables should be loaded (and the order
    they should appear in the output) from the given sequences of
    ``variables`` and ``categories``.

    If neither ``variables`` nor ``categories`` are provided, ``None`` is
    returned, indicating that all variables should be loaded.

    :arg cattable:   The category table
    :arg variables:  List of variable IDs to import. May be ``None``.
    :arg categories: List of category names to import. May be ``None``.
    :returns:        Sequence of variables to load, or ``None`` if all
                     variables should be loaded.
    """

    # Build a list of all the variables we
    # want to load, from the variables and
    # categories that were passed in.
    if categories is not None:

        if variables is None:
            variables = []

        catvars   = loadtables.categoryVariables(cattable, categories)
        variables = variables + [c for c in catvars if c not in variables]

    return variables


def columnsToLoad(datafiles,
                  vartable,
                  variables,
                  colnames,
                  unknownVars,
                  removeUnknown,
                  indexes=None,
                  sniffers=None):
    """Determines which columns should be loaded from ``datafiles``.

    Peeks at the first line of the data file (assumed to contain column names),
    then uses the variable table to determine which of them should be loaded.

    :arg datafiles:     Path to data file(s)

    :arg vartable:      Variable table

    :arg variables:     List of variables to load. If provided,
                        ``removeUnknown`` is ignored.

    :arg colnames:      List of column names/glob-style wildcard patterns,
                        specifying columns to load. If provided,
                        ``removeUnknown`` is ignored.

    :arg unknownVars:   List of :class:`.Column` objects representing unknown
                        variables

    :arg removeUnknown: If ``True``, any variables in ``datafile`` which are
                        not in ``vartable`` are not loaded.

    :arg indexes:       Dict of ``{filename : [index]}`` mappings, specifying
                        the position of the column(s) to use as the index.
                        Defaults to 0 (the first column).

    :arg sniffers:      Dict of ``{ file : snifferName }`` mappings containing
                        custom sniffers to be used for specific files. See the
                        :mod:`.custom` module.

    :returns:           A tuple containing:

                         - A dict of ``{ file : [Column] }`` mappings, the
                           :class:`.Column` objects to *load* from each input
                           file. The columns (including the index column) are
                           ordered as they appear in the file.

                         - A list containing the :class:`.Column` objects to
                           *ignore*.
    """

    if sniffers    is     None: sniffers      = {}
    if indexes     is     None: indexes       = {}
    if unknownVars is     None: unknownVars   = []
    if variables   is not None: removeUnknown = False
    if colnames    is not None: removeUnknown = False

    # Turn the unknonwVars list
    # into a list of variable IDs
    unknownVids = list(sorted({c.vid for c in unknownVars}))

    if isinstance(datafiles, str):
        datafiles = [datafiles]

    # We apply these cleaning steps by
    # omitting the relevant columns.
    loadFuncNames = ['remove', 'keepVisits']

    # Peek at the columns that
    # are in the input files.
    allcols = fileinfo.fileinfo(datafiles,
                                indexes=indexes,
                                sniffers=sniffers)[2]
    ncols   = len(list(it.chain(*allcols)))

    # re-organise the columns - a list of
    # columns for each variable ID. We do
    # this because, for a given VID, we
    # want to pass all columns at once to
    # the cleaning function(s) below.
    byvid = collections.defaultdict(list)
    for col in it.chain(*allcols):
        byvid[col.vid].append(col)

    # Build a full list of index
    # columns for each data file.
    indexes = [indexes.get(f, [0]) for f in datafiles]

    # retrieve all cleaning steps -
    # we are only going to apply the
    # cleaning steps that will
    # determine whether or not a column
    # should be loaded
    mask    = vartable['Clean'].notna()
    cleans  = vartable['Clean'][mask]
    ppvids  = vartable.index[   mask]

    # Loop through all columns in
    # the data, and build a list of
    # the ones we want to load. The
    # end result will be an ordered
    # dict of { file : [column] }
    # mappings, and a list of columns
    # to drop.
    drop = []
    load = collections.OrderedDict([(f, []) for f in datafiles])
    for vid, cols in byvid.items():

        # index column - load it!
        # (the fileinfo function gives
        # index columns a variable ID
        # of 0).
        if vid == 0:
            for col in cols:
                load[col.datafile].append(col)
            continue

        # Variable is flagged as unknown,
        # and we have been told to ignore
        # unknown variables. Remember that
        # removeUnknown has no effect if
        # variables/column patterns have
        # been specified.
        if removeUnknown and vid in unknownVids:
            drop.extend(cols)
            continue

        # Figure out whether each
        # column should be loaded.
        # We load all columns which
        # pass either the variables
        # test or the colnames test
        # (or, if neither of those
        # options have been given,
        # all columns)
        loadflags = [(variables is None) and (colnames is None) for c in cols]

        # variable list has been specified,
        # and this vid is not in it - don't
        # load.
        if variables is not None:
            loadflags = [(vid in variables) for c in cols]

        # column names/patterns specified -
        # filter the list of columns based
        # on whether they match any of the
        # patterns specified.
        if colnames is not None:
            for i, col in enumerate(cols):
                hits = [fnmatch.fnmatch(col.name, pat) for pat in colnames]
                loadflags[i] = loadflags[i] or any(hits)

        for col, loadflag in list(zip(cols, loadflags)):
            if not loadflag:
                cols.remove(col)
                drop.append(col)

        if len(cols) == 0:
            continue

        # cleaning specified for this variable
        if vid in ppvids:

            # retrieve the cleaning functions
            # which affect whether or not a column
            # should get loaded. We remove these
            # functions from the variable table, as
            # they won't need to be called again.
            funcs = [cleans[vid].pop(n, None) for n in loadFuncNames]
            funcs = [f for f in funcs if f is not None]

            # call the functions, generate a new
            # set of columns for this variable
            newcols = cols
            for f in funcs:
                newcols = f.run(vartable, vid, newcols)

            drop.extend(list(set.difference(set(cols), set(newcols))))

            cols = newcols

        for col in cols:
            load[col.datafile].append(col)

    # Final step - the column lists for each
    # file are not necessarily ordered by
    # their position in the file. Re-order
    # them so they are.
    for fname, cols in list(load.items()):
        load[fname].sort(key=lambda c: c.index)

    log.debug('Identified %i / %i columns to be loaded',
              sum([len(c) for c in load.values()]), ncols)

    return load, drop


def filterSubjects(data,
                   cols,
                   subjects=None,
                   subjectExprs=None,
                   exclude=None):
    """Removes rows (subjects) from the data based on ``subjects`` to
    include, conditional ``subjectExprs``, and subjects to ``exclude``.

    :arg data:         A ``pandas.DataFrame`` instance.

    :arg allcols:      List of :class:`.Column` objects describing every column
                       in the data set.

    :arg subjects:     List of subjects to include.

    :arg subjectExprs: List of subject inclusion expressions

    :arg exclude:      List of subjects to exclude

    :returns:          A ``pandas.DataFrame``, potentially with rows removed.
    """

    if all((subjects     is None,
            subjectExprs is None,
            exclude      is None)):
        return data

    mask = None

    # ones to include, zeros to drop
    if subjects is not None:
        mask = data.index.isin(subjects)

    if subjectExprs is not None and len(subjectExprs) >= 1:
        exprmask = evaluateSubjectExpressions(data, cols, subjectExprs)
        # include rows listed in subjects
        # and which pass any expression
        if mask is not None: mask = mask & exprmask
        else:                mask = exprmask

    # exclude list overrides all of the above
    if exclude is not None:
        exclmask = data.index.isin(exclude)
        if mask is None: mask           = ~exclmask
        else:            mask[exclmask] = 0

    if mask is not None: return data[mask]
    else:                return data


def evaluateSubjectExpressions(data, allcols, subjectExprs):
    """Remove subjects (rows) from the data according to ``subjectExprs``.

    :arg data:         A ``pandas.DataFrame`` instance.

    :arg allcols:      List of :class:`.Column` objects describing every column
                       in the data set.

    :arg subjectExprs: List of strings containing expressions which identify
                       subjects to be included. Subjects for which *any*
                       expression evaluates to ``True`` will be included.

    :returns:          1D boolean ``numpy`` array containing ``True`` for
                       subjects to be included and ``False`` for subjects to
                       be excluded. Or ``None``, indicating that the
                       expressions were not evaluated (and all rows passed).
    """

    # build a {vid : [column]} mapping
    # to make life easy for the
    # evaluateSubjectExpression function
    colsbyvid = collections.defaultdict(list)
    for col in allcols:
        colsbyvid[col.vid].append(col)

    # evaluate each expression - we get
    # a numpy array for each of them
    exprmasks = []
    for i, expr in enumerate(subjectExprs):
        exprmask = evaluateSubjectExpression(data, expr, colsbyvid)
        if exprmask is not None:
            exprmasks.append(exprmask)

    # Any result which was not combined using
    # any() or all() defaults to being combined
    # with any(). For example, if "v123 >= 2"
    # is applied to columns 123-0.0, 123-1.0,
    # and 123-2.0, the final result will be
    # a 1D boolean array containing True where
    # any of the three columns were >= 2.
    for i, em in enumerate(exprmasks):
        if len(em.shape) == 2:
            exprmasks[i] = em.any(axis=1)

    # Finally, all expressions are combined in
    # the same manner - i.e. rows which passed
    # *any* of the expressions are included
    if   len(exprmasks) >  1: return ft.reduce(lambda a, b: a | b, exprmasks)
    elif len(exprmasks) == 1: return exprmasks[0]
    else:                     return None


def evaluateSubjectExpression(data, expr, cols):
    """Evaluates the given variable expression for each row in the data.

    :arg data: A ``pandas.DataFrame`` instance.

    :arg expr: String containing a variable expression

    :arg cols: Dict of ``{vid : [Column]}`` mappings

    :returns:  A boolean ``numpy`` array containing the result of evaluating
               the expression at each row, or ``None`` indicating that the
               expression was not evaluated (and every row passed).
    """

    expr = expression.Expression(expr)
    vids = expr.variables

    # Calculate the intersection of visits/
    # instances across all variables used in
    # the expression - we evaluate an
    # expression only on visits/instances
    # present for all variables in the
    # expression. All other visits/instances
    # are not considered.
    visits    = [[c.visit    for c in cols[v]] for v in vids]
    instances = [[c.instance for c in cols[v]] for v in vids]

    def intersection(a, b):
        return set(a).intersection(b)
    intersection = ft.partial(ft.reduce, intersection)

    visits    = intersection(visits)
    instances = intersection(instances)

    # Build a {vid : [column]} dict to pass
    # to the expression evaluate method
    exprcols = collections.defaultdict(list)
    for vid in vids:
        for col in cols[vid]:
            if col.visit in visits and col.instance in instances:
                exprcols[vid].append(col.name)

    if len(exprcols) == 0:
        log.debug('Ignoring expression (%s) - no associated '
                  'columns are present', str(expr))
        return None

    log.debug('Evaluating expression (%s) on columns %s',
              str(expr), list(it.chain(*exprcols.values())))

    return expr.evaluate(data, exprcols)



#!/usr/bin/env python
#
# test_cleaning.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import warnings
import random
import string
import textwrap as tw

import multiprocessing as mp
from unittest import mock

import pytest

import numpy  as np
import pandas as pd

import funpack.loadtables as lt
import funpack.cleaning as cleaning
import funpack.custom as custom
import funpack.cleaning_functions as cfns

import funpack.importing as importing
import funpack.hierarchy as hierarchy

from . import tempdir, gen_test_data, gen_tables, clear_plugins


def test_remove():
    assert cfns.remove(None, 0, ['a', 'b', 'c']) == []


def test_keepVisits():
    # first,
    # last,
    # number

    with tempdir():

        gen_test_data(20, 10, 'data.txt', max_visits=10, start_var=50)
        vartable, proctable, cattable, _ = gen_tables(range(50, 60))

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False)

        # find a column with several visits
        for v in dt.variables:
            cols = dt.columns(v)
            if len(cols) > 5:
                break

        assert cfns.keepVisits(vartable, v, cols, 'first') == [cols[0]]
        assert cfns.keepVisits(vartable, v, cols, 'last')  == [cols[-1]]
        assert cfns.keepVisits(vartable, v, cols, 2)       == [cols[2]]
        assert cfns.keepVisits(vartable, v, cols, 3, 4)    == cols[3:5]


def test_keepVisits_not_instancing_2():

    data = tw.dedent("""
    eid\t1-0.0\t1-1.0\t1-2.0\t2-0.0\t2-1.0\t2-2.0
    1\t10\t11\t12\t20\t21\t22
    2\t10\t11\t12\t20\t21\t22
    3\t10\t11\t12\t20\t21\t22
    4\t10\t11\t12\t20\t21\t22
    5\t10\t11\t12\t20\t21\t22
    """).strip()
    with tempdir():

        with open('data.tsv', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables([1, 2])

        vartable.loc[1, 'Instancing'] = 2
        vartable.loc[2, 'Instancing'] = 0

        dt, _ = importing.importData('data.tsv', vartable, proctable, cattable)

        v1cols = dt.columns(1)
        v2cols = dt.columns(2)

        assert cfns.keepVisits(dt.vartable, 1, v1cols, 'first') == [v1cols[0]]
        assert cfns.keepVisits(dt.vartable, 1, v1cols, 'last')  == [v1cols[-1]]
        assert cfns.keepVisits(dt.vartable, 1, v1cols, 0, 1)    ==  v1cols[:2]
        assert cfns.keepVisits(dt.vartable, 1, v1cols, 1)       == [v1cols[1]]

        # v2 does not have instancing 2, should be ignored
        assert cfns.keepVisits(dt.vartable, 2, v2cols, 'first') == v2cols
        assert cfns.keepVisits(dt.vartable, 2, v2cols, 'last')  == v2cols
        assert cfns.keepVisits(dt.vartable, 2, v2cols, 0)       == v2cols
        assert cfns.keepVisits(dt.vartable, 2, v2cols, 2)       == v2cols
        assert cfns.keepVisits(dt.vartable, 2, v2cols, 1)       == v2cols
        assert cfns.keepVisits(dt.vartable, 2, v2cols, 0, 1)    == v2cols


def test_fillVisits():

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()

        data        = np.zeros((20, 4), dtype=np.float32)
        data[:, 0]  = np.arange(1, 21)
        data[:, 1:] = np.random.randint(1, 100, (20, 3))

        cols = ['eid', '1-0.0', '1-1.0', '1-2.0']

        data[ 0:10, 1] = np.nan
        data[ 5:15, 2] = np.nan
        data[10:,   3] = np.nan

        meanexp           = np.copy(data)
        meanexp[ 0:5,  1] = np.mean(data[0:5, 2:4], axis=1)
        meanexp[ 5:10, 1] = data[ 5:10, 3]
        meanexp[ 5:10, 2] = data[ 5:10, 3]
        meanexp[10:15, 2] = data[10:15, 1]
        meanexp[10:15, 3] = data[10:15, 1]
        meanexp[15:20, 3] = np.mean(data[15:20, 1:3], axis=1)

        modeexp           = np.copy(meanexp)

        m1 = pd.DataFrame({'a' : data[ 0:5,  2], 'b' : data[ 0:5,  3]})
        m1 = np.asarray(m1.mode(axis=1).iloc[:, 0])

        m2 = pd.DataFrame({'a' : data[15:20, 1], 'b' : data[15:20, 2]})
        m2 = np.asarray(m2.mode(axis=1).iloc[:, 0])

        modeexp[ 0:5,  1] = m1
        modeexp[15:20, 3] = m2

        with open('data.txt', 'wt') as f:

            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')

        vartable, proctable, cattable, _ = gen_tables([1])

        for method in ['mean', 'mode']:

            dt, _ = importing.importData('data.txt',
                                         vartable,
                                         proctable,
                                         cattable,
                                         removeUnknown=False,
                                         pool=pool,
                                         mgr=mgr)

            cfns.fillVisits(dt, 1, method)

            if   method == 'mean': exp = meanexp
            elif method == 'mode': exp = modeexp

            assert np.all(dt[:, '1-0.0'] == exp[:, 1])
            assert np.all(dt[:, '1-1.0'] == exp[:, 2])
            assert np.all(dt[:, '1-2.0'] == exp[:, 3])

        with pytest.raises(ValueError):
            cfns.fillVisits(dt, 1, 'bad_method')
        dt = None
        mgr = None
        pool = None



def test_fillMissing():
    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()

        data        = np.zeros((20, 3), dtype=np.float32)
        data[:, 0]  = np.arange(1, 21)
        data[:, 1:] = np.random.randint(1, 100, (20, 2))

        cols = ['eid', '1-0.0', '2-0.0']

        data[ 0:10, 1] = np.nan
        data[ 5:15, 2] = np.nan

        exp = np.copy(data)
        exp[ 0:10, 1] = 999
        exp[ 5:15, 2] = 12345

        with open('data.txt', 'wt') as f:

            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')

        vartable, proctable, cattable, _ = gen_tables([1])

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        cfns.fillMissing(dt, 1, 999)
        cfns.fillMissing(dt, 2, 12345)

        assert np.all(dt[:, '1-0.0'] == exp[:, 1])
        assert np.all(dt[:, '2-0.0'] == exp[:, 2])
        dt = None
        mgr = None
        pool = None


def test_codeToNumeric():

    hier  = hierarchy.getHierarchyFilePath(name='opcs3')
    hier  = hierarchy.loadHierarchyFile(hier)
    codes = list(random.sample(hier.codings, 20)) + ['badcode']
    exp   = [hier.index(c) + 1 for c in codes[:-1]] + [np.nan]

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write('eid\t1-0.0\n')
            for i, c in enumerate(codes, 1):
                f.write('{}\t{}\n'.format(i, c))

        vartable, proctable, cattable, _ = gen_tables([1])

        vartable.loc[1, 'DataCoding'] = hierarchy.HIERARCHY_DATA_NAMES['opcs3']

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False)
        cfns.codeToNumeric(dt, 1)

        got = dt[:, '1-0.0']
        exp = pd.Series(exp, name='1-0.0', index=got.index)

        gotna = pd.isna(got)
        expna = pd.isna(exp)

        assert np.all(got[~gotna] == exp[~expna])
        assert np.all(gotna       == expna)
        dt = None



def test_convertICD10Codes():

    icd10hier = hierarchy.getHierarchyFilePath(name='icd10')
    icd10hier = hierarchy.loadHierarchyFile(icd10hier)
    codes     = list(random.sample(icd10hier.codings, 20)) + ['badcode']
    exp       = [icd10hier.index(c) + 1 for c in codes[:-1]] + [np.nan]

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:
        mgr = mp.Manager()
        with open('data.txt', 'wt') as f:
            f.write('eid\t1-0.0\n')
            for i, c in enumerate(codes, 1):
                f.write('{}\t{}\n'.format(i, c))

        vartable, proctable, cattable, _ = gen_tables([1])
        vartable.loc[1, 'DataCoding'] = hierarchy.HIERARCHY_DATA_NAMES['icd10']

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)
        cfns.convertICD10Codes(dt, 1)

        got = dt[:, '1-0.0']
        exp = pd.Series(exp, name='1-0.0', index=got.index)

        gotna = pd.isna(got)
        expna = pd.isna(exp)

        assert np.all(got[~gotna] == exp[~expna])
        assert np.all(gotna       == expna)
        dt = None
        mgr = None
        pool = None


def test_makeNa():

    data        = np.zeros((100, 3), dtype=np.float32)
    data[:, 0]  = np.arange(1, 101)
    data[:, 1:] = np.random.randint(1, 100, (100, 2))

    cols = ['eid', '1-0.0', '2-0.0']

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()
        with open('data.txt', 'wt') as f:
            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')
        vartable, proctable, cattable, _ = gen_tables([1, 2])

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        cfns.makeNa(dt, 1, '>= 25')
        cfns.makeNa(dt, 2, '<  50')

        na1mask = data[:, 1] >= 25
        na2mask = data[:, 2] <  50

        dt1 = dt[dt[:, '1-0.0'].notna(), '1-0.0']
        dt2 = dt[dt[:, '2-0.0'].notna(), '2-0.0']

        assert np.all(      na1mask     == dt[:, '1-0.0'].isna())
        assert np.all(      na2mask     == dt[:, '2-0.0'].isna())
        assert np.all(data[~na1mask, 1] == dt1)
        assert np.all(data[~na2mask, 2] == dt2)
        dt = None
        mgr = None
        pool = None


def test_applyNewLevels():

    data       = np.random.randint(0, 10, (50, 4))
    data[:, 0] = np.arange(1, 51)

    codes1 = np.random.randint(100, 200, 10)
    codes2 = np.random.randint(100, 200, 10)
    codes3 = list(reversed(range(10)))

    exp1 = [codes1[data[i, 1]] for i in range(50)]
    exp2 = [codes2[data[i, 2]] for i in range(50)]
    exp3 = [codes3[data[i, 3]] for i in range(50)]

    vartable, proctable, cattable, _ = gen_tables([1, 2, 3])

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        vartable['RawLevels'][1] = list(range(10))
        vartable['NewLevels'][1] = codes1
        vartable['RawLevels'][2] = list(range(10))
        vartable['NewLevels'][2] = codes2
        vartable['RawLevels'][3] = list(range(10))
        vartable['NewLevels'][3] = codes3


    cols = ['eid', '1-0.0', '2-0.0', '3-0.0']

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False)

        cleaning.applyNewLevels(dt)

        assert np.all(dt[:, '1-0.0'] == exp1)
        assert np.all(dt[:, '2-0.0'] == exp2)
        assert np.all(dt[:, '3-0.0'] == exp3)

        col3 = dt.columns(3)[0]

        assert dt.getFlags(col3) == set(['inverted'])
        dt = None


def test_applyNAInsertion():

    data       = np.random.randint(0, 10, (100, 3)).astype(np.float32)
    data[:, 0] = np.arange(1, 101)

    miss1 = list(np.random.choice(range(10), 4, replace=False))
    miss2 = list(np.random.choice(range(10), 4, replace=False))

    exp1 = data[:, 1].copy()
    exp2 = data[:, 2].copy()

    for m in miss1: exp1[exp1 == m] = np.nan
    for m in miss2: exp2[exp2 == m] = np.nan

    vartable, proctable, cattable, _ = gen_tables([1, 2, 3])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        vartable['NAValues'][1] = miss1
        vartable['NAValues'][2] = miss2
        vartable['NAValues'][3] = [1, 2, 3]

    cols = ['eid', '1-0.0', '2-0.0']

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()
        with open('data.txt', 'wt') as f:
            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        cleaning.applyNAInsertion(dt)

        na1mask = dt[:, '1-0.0'].isna()
        na2mask = dt[:, '2-0.0'].isna()

        d1 = dt[:, '1-0.0'][~na1mask]
        d2 = dt[:, '2-0.0'][~na2mask]

        assert np.all(na1mask == np.isnan(exp1))
        assert np.all(na2mask == np.isnan(exp2))

        assert np.all(d1 == exp1[~na1mask])
        assert np.all(d2 == exp2[~na2mask])
        dt = None
        mgr = None
        pool = None


def test_applyChildValues():

    sz          = 100
    data        = np.zeros((sz, 6), dtype=np.float32)
    data[:, 0]  = np.arange(1, sz + 1)
    data[:, 1:] = np.random.randint(1, 10, (sz, 5))

    cols = ['eid', '1-0.0', '2-0.0', '3-0.0', '4-0.0', '5-0.0']

    # parents
    # 1: 2, 3
    # 2: 3
    # 3: 4
    # 4: 5

    pvals = {
        1 : 'v2 == 5, v3 > 5',
        2 : 'v3 < 8',
        3 : 'v4 >= 6',
        4 : 'v5 == 2',
        5 : 'v123 == 7',
        6 : 'v8 < 5'
    }
    cvals = {
        1 : '100, 101',
        2 : '200',
        3 : '300',
        4 : '400',
        5 : '123',
        6 : '1234'
    }

    data[ data[:, 5] == 2,                     4] = np.nan
    data[ data[:, 4] >= 6,                     3] = np.nan
    data[ data[:, 3] <  8,                     2] = np.nan
    data[(data[:, 2] == 5) | (data[:, 3] > 5), 1] = np.nan

    exp  = np.copy(data)
    nan1 = np.isnan(exp[:, 1])
    exp[    nan1            & (exp[:, 2] == 5), 1] = 100
    exp[    nan1            & (exp[:, 3] >  5), 1] = 101
    exp[np.isnan(exp[:, 2]) & (exp[:, 3] <  8), 2] = 200
    exp[np.isnan(exp[:, 3]) & (exp[:, 4] >= 6), 3] = 300
    exp[np.isnan(exp[:, 4]) & (exp[:, 5] == 2), 4] = 400

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()
        with open('data.txt', 'wt') as f:
            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')

        vartable, proctable, cattable, _ = gen_tables([1, 2, 3, 4, 5, 6])

        vartable.loc[pvals.keys(), 'ParentValues'] = \
            [lt.convert_ParentValues(v) for v in pvals.values()]
        vartable.loc[cvals.keys(), 'ChildValues'] = \
            [lt.convert_comma_sep_numbers(v) for v in cvals.values()]

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        cleaning.applyChildValues(dt)

        assert np.all(np.asarray(dt[:, :].values) == exp[:, 1:])
        dt = None
        mgr = None
        pool = None


@clear_plugins
def test_applyCleaningFunctions():

    data        = np.zeros((100, 3), dtype=np.float32)
    data[:, 0]  = np.arange(1, 101)
    data[:, 1:] = np.random.randint(1, 100, (100, 2))

    cols = ['eid', '1-0.0', '2-0.0']

    @custom.cleaner()
    def proc1(dtable, vid):
        columns = dtable.columns(vid)
        for c in columns:
            dtable[:, c.name] = dtable[:, c.name] + 5


    @custom.cleaner()
    def proc2(dtable, vid, factor):
        columns = dtable.columns(vid)
        for c in columns:
            dtable[:, c.name] = dtable[:, c.name] * factor


    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()
        with open('data.txt', 'wt') as f:
            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')
        vartable, proctable, cattable, _ = gen_tables([1, 2, 3])

        vartable.loc[(1, ), 'Clean'] = (lt.convert_Process(
            'cleaner', 'proc1'), )
        vartable.loc[(2, ), 'Clean'] = (lt.convert_Process(
            'cleaner', 'proc2(50)'), )
        vartable.loc[(3, ), 'Clean'] = (lt.convert_Process(
            'cleaner', 'proc2(50)'), )

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        cleaning.applyCleaningFunctions(dt)

        assert np.all(dt[:, '1-0.0'] == data[:, 1] + 5)
        assert np.all(dt[:, '2-0.0'] == data[:, 2] * 50)
        dt = None
        mgr = None
        pool = None


def test_cleanData():

    data        = np.zeros((100, 3), dtype=np.float32)
    data[:, 0]  = np.arange(1, 101)
    data[:, 1:] = np.random.randint(1, 100, (100, 2))

    cols = ['eid', '1-0.0', '2-0.0']

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        mgr = mp.Manager()
        custom.registerBuiltIns()
        with open('data.txt', 'wt') as f:
            f.write('\t'.join(cols) + '\n')
            np.savetxt(f, data, delimiter='\t')
        vartable, proctable, cattable, _ = gen_tables([1, 2])
        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        # not crashing -> pass
        cleaning.cleanData(dt)
        dt = None
        mgr = None
        pool = None


def test_parseSpirometryData():

    data = tw.dedent("""
    eid\t1-0.0\t1-0.1\t1-0.2
    1\tblow1,5,1,2,3,4,5\tblow2,3,4,5,6\tblow3,6,1,2,3,4,5,6
    2\tblow2,5,1,2,3,4,5\tblow3,5,5,4,3,2,1\tblow1,3,3,2,1
    3\tblow1,5,1,2,3,4,5\tblow3,4,6,7,8,9\tblow2,2,1,2
    """).strip()

    # arranged by column
    exp = [
        [[1, 2, 3, 4, 5],    [3, 2, 1],       [1, 2, 3, 4, 5]],
        [[4, 5, 6],          [1, 2, 3, 4, 5], [1, 2]],
        [[1, 2, 3, 4, 5, 6], [5, 4, 3, 2, 1], [6, 7, 8, 9]]]

    exp = [[np.array(row) for row in col] for col in exp]
    exp = [pd.Series(col) for col in exp]

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:
        with open('data.txt', 'wt') as f:
            f.write(data)

        mgr = mp.Manager()
        vartable, proctable, cattable, _ = gen_tables([1, 2])

        dt, _ = importing.importData('data.txt',
                                     vartable,
                                     proctable,
                                     cattable,
                                     removeUnknown=False,
                                     pool=pool,
                                     mgr=mgr)

        cfns.parseSpirometryData(dt, 1)

        for coli, col in enumerate(['1-0.0', '1-0.1', '1-0.2']):
            got = dt[:, col]

            for rowi in range(len(got)):
                assert np.all(exp[coli][rowi] == got.iloc[rowi])
        dt = None
        mgr = None
        pool = None


def test_flattenHierarchical():

    data = tw.dedent("""
    eid,1-0.0
    1,10
    2,20
    3,30
    4,40
    5,50
    6,60
    7,70
    8,80
    9,90
    10,
    """).strip()

    hier = tw.dedent("""
    coding	meaning	node_id	parent_id
    10	meaning 10	1	0
    20	meaning 20	2	0
    30	meaning 30	3	1
    40	meaning 40	4	1
    50	meaning 50	5	4
    60	meaning 60	6	2
    70	meaning 70	7	2
    80	meaning 80	8	6
    90	meaning 90	9	8
    """).strip()

    tests = [
        (0, [10, 20, 10, 10, 10, 20, 20, 20, 20]),
        (1, [10, 20, 30, 40, 40, 60, 70, 60, 60]),
        (2, [10, 20, 30, 40, 50, 60, 70, 80, 80]),
        (3, [10, 20, 30, 40, 50, 60, 70, 80, 90]),
        (4, [10, 20, 30, 40, 50, 60, 70, 80, 90]),
    ]

    with tempdir(), mp.Pool(mp.cpu_count()) as pool:

        with open('data.txt', 'wt') as f:
            f.write(data)

        with open('hier.txt', 'wt') as f:
            f.write(hier)

        mgr = mp.Manager()
        vartable, proctable, cattable, _ = gen_tables([1])

        for level, exp in tests:

            dt, _ = importing.importData('data.txt',
                                         vartable,
                                         proctable,
                                         cattable,
                                         removeUnknown=False,
                                         pool=pool,
                                         mgr=mgr)

            with mock.patch('funpack.hierarchy.getHierarchyFilePath',
                            return_value='hier.txt'):
                cfns.flattenHierarchical(dt, 1, level)

            vals = dt[:, '1-0.0'].values

            assert pd.isna(vals[ -1])
            assert np.all( vals[:-1] == exp)

        dt = None
        mgr = None
        pool = None



#!/usr/bin/env python

import textwrap as tw

import numpy  as np
import pandas as pd

import pytest

from . import tempdir

import funpack.main   as main
import funpack.custom as custom
import funpack.config as config


def test_InternalType_trustTypes():
    _test_InternalType(True)

def test_InternalType_no_trustTypes():
    _test_InternalType(False)


def _test_InternalType(trustTypes):


    data = tw.dedent("""
    eid\t1-0.0\t2-0.0\t3-0.0\t4-0.0\t5-0.0
    1\t1000\t2\t3\t4\t5
    2\t1\t2\t3\t4\t5
    3\t1\t2\t3\t4\t5
    4\t1\t2\t3\t4\t5
    5\t1\t2\t3\t4\t5
    """).strip()

    vartable = tw.dedent("""
    ID\tInternalType
    1\tint8
    2\tfloat64
    3\tuint32
    4\tfloat32
    5\tint16
    """).strip()

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)
        with open('vartable.txt', 'wt') as f:
            f.write(vartable)

        custom.registerBuiltIns()
        args   = '-nb -vf vartable.txt out.txt data.txt'
        if trustTypes:
            args += ' -tt'
        args   = config.parseArgs(args.split())[0]
        dtable = main.doImport(args, None, None)[0]

        assert dtable[:, '1-0.0'].dtype == np.int8
        assert dtable[:, '2-0.0'].dtype == np.float64
        assert dtable[:, '3-0.0'].dtype == np.uint32
        assert dtable[:, '4-0.0'].dtype == np.float32
        assert dtable[:, '5-0.0'].dtype == np.int16

        # The low 8 bits of 1000 == -24
        # when interpreted as int8
        assert dtable[1, '1-0.0'] == -24



def test_trustTypes():

    data = tw.dedent("""
    eid\t1-0.0
    1\t1
    2\t1
    3\tbadval
    """).strip()

    vartable = tw.dedent("""
    ID\tInternalType
    1\tfloat32
    """).strip()

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)
        with open('vartable.txt', 'wt') as f:
            f.write(vartable)

        custom.registerBuiltIns()

        args = '-nb -tt -vf vartable.txt out.txt data.txt'
        args = config.parseArgs(args.split())[0]

        with pytest.raises(ValueError):
            dtable = main.doImport(args, None, None)[0]

        args   = '-nb -vf vartable.txt out.txt data.txt'
        args   = config.parseArgs(args.split())[0]
        dtable = main.doImport(args, None, None)[0]

        assert dtable[1, '1-0.0'] == 1
        assert dtable[2, '1-0.0'] == 1
        assert pd.isna(dtable[3, '1-0.0'])


def test_largevals():

    # Some random values which cannot be
    # represented with float32 data type
    vals = np.array([10001249182,  48192124975, 14328901242])

    data = 'eid\t1-0.0\n'
    for i, v in enumerate(vals):
        data += '{}\t{}\n'.format(i, v)



    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)

        # If we specify the variable as an integer,
        # funpack will default to float32, and the
        # values will be misinterpreted
        vartable = 'ID\tType\n1\tinteger'
        with open('vartable.txt', 'wt') as f:
            f.write(vartable)

        custom.registerBuiltIns()

        args = '-nb -vf vartable.txt out.txt data.txt'
        args = config.parseArgs(args.split())[0]
        dtable = main.doImport(args, None, None)[0]

        assert not np.any(dtable[:, '1-0.0'] == vals)

        # But we can use InternalType to
        # specify a suitable storage type
        vartable = 'ID\tType\tInternalType\n1\tinteger\tfloat64'
        with open('vartable.txt', 'wt') as f:
            f.write(vartable)

        args   = '-nb -vf vartable.txt out.txt data.txt'
        args   = config.parseArgs(args.split())[0]
        dtable = main.doImport(args, None, None)[0]

        assert np.all(dtable[:, '1-0.0'] == vals)
